# (C) 2025 GoodData Corporation

"""Module for provisioning user data filters in GoodData workspaces.

This module provides the `UserDataFilterProvisioner` class, which is responsible
for creating, updating, and deleting user data filters in GoodData workspaces.
"""

import re

from gooddata_sdk.catalog.workspace.entity_model.user_data_filter import (
    CatalogEntityIdentifier,
    CatalogUserDataFilter,
    CatalogUserDataFilterAttributes,
    CatalogUserDataFilterRelationships,
)

from gooddata_pipelines.provisioning.entities.user_data_filters.models.udf_models import (
    UserDataFilterFullLoad,
    UserDataFilterGroup,
    UserDataFilterIncrementalLoad,
    WorkspaceUserDataFilters,
)
from gooddata_pipelines.provisioning.provisioning import Provisioning
from gooddata_pipelines.provisioning.utils.exceptions import ContextException


class UserDataFilterProvisioner(
    Provisioning[UserDataFilterFullLoad, UserDataFilterIncrementalLoad]
):
    """Provisioning class for user data filters in GoodData workspaces.

    This class handles the creation, update, and deletion of user data filters
    based on the provided source data.

    Requires setting the `ldm_column_name` and `maql_column_name`
    attributes before calling the `provision` method.

    Usage:
    ```
    provisioner = UserDataFilterProvisioner(api, source_group)
    provisioner.set_ldm_column_name("ldm_column")
    provisioner.set_maql_column_name("maql_column")
    provisioner.provision()
    ```
    """

    source_group_full: list[UserDataFilterFullLoad]
    source_group_incremental: list[UserDataFilterIncrementalLoad]
    ldm_column_name: str = ""
    maql_column_name: str = ""

    FULL_LOAD_TYPE = UserDataFilterFullLoad

    def set_ldm_column_name(self, ldm_column_name: str) -> None:
        """Set the LDM column name for user data filters.

        Args:
            ldm_column_name (str): The LDM column name to set.
        """
        self.ldm_column_name = ldm_column_name

    def set_maql_column_name(self, maql_column_name: str) -> None:
        """Set the MAQL column name for user data filters.

        Args:
            maql_column_name (str): The MAQL column name to set.
        """
        self.maql_column_name = maql_column_name

    @staticmethod
    def _group_db_user_data_filters_by_ws_id(
        user_data_filters: list[UserDataFilterFullLoad],
    ) -> list[WorkspaceUserDataFilters]:
        """Group user data filters by workspace ID and user ID."""
        ws_map: dict[str, dict[str, set[str]]] = {}

        for udf in user_data_filters:
            ws_map.setdefault(udf.workspace_id, {}).setdefault(
                udf.udf_id, set()
            ).add(str(udf.udf_value))

        result: list[WorkspaceUserDataFilters] = []

        for ws_id, udf_dict in ws_map.items():
            udf_groups = [
                UserDataFilterGroup(udf_id=udf_id, udf_values=list(values))
                for udf_id, values in udf_dict.items()
            ]
            result.append(
                WorkspaceUserDataFilters(
                    workspace_id=ws_id, user_data_filters=udf_groups
                )
            )
        return result

    @staticmethod
    def _extract_numbers_from_maql(maql: str) -> list[str]:
        """Extract numbers from a MAQL string."""
        numbers = re.findall(r'"\d+"', maql)
        return [number.strip('"') for number in numbers]

    def _skip_user_data_filter_update(
        self, existing_udf: list[CatalogUserDataFilter], udf_value: list[str]
    ) -> bool:
        """Check if the user data filter update can be skipped."""
        if not existing_udf:
            return False
        existing_udfs = self._extract_numbers_from_maql(
            existing_udf[0].attributes.maql
        )
        return set(udf_value) == set(existing_udfs)

    def _create_user_data_filters(
        self, user_data_filter_ids_to_create: list[WorkspaceUserDataFilters]
    ) -> None:
        """Create or update user data filters in GoodData workspaces."""
        for workspace_user_data_filter in user_data_filter_ids_to_create:
            workspace_id = workspace_user_data_filter.workspace_id
            user_data_filters = workspace_user_data_filter.user_data_filters

            gd_user_data_filters: list[CatalogUserDataFilter] = (
                self._api.list_user_data_filters(workspace_id)
            )

            gd_udf_ids = {
                user.relationships.user["data"].id
                for user in gd_user_data_filters
                if user.relationships and user.relationships.user
            }

            db_udf_ids = {udf.udf_id for udf in user_data_filters}

            udf_ids_to_delete: set[str] = gd_udf_ids.difference(db_udf_ids)
            self._delete_user_data_filters(workspace_id, udf_ids_to_delete)

            udf_group: UserDataFilterGroup
            for udf_group in user_data_filters:
                udf_id: str = udf_group.udf_id
                udf_values: list[str] = udf_group.udf_values

                existing_udf: list[CatalogUserDataFilter] = [
                    udf for udf in gd_user_data_filters if udf.id == udf_id
                ]
                if self._skip_user_data_filter_update(existing_udf, udf_values):
                    continue

                formatted_udf_values = '", "'.join(
                    str(value) for value in udf_values
                )
                maql = f'{self.maql_column_name} IN ("{formatted_udf_values}")'

                attributes = CatalogUserDataFilterAttributes(maql=maql)
                relationships = CatalogUserDataFilterRelationships(
                    labels={
                        "data": [
                            CatalogEntityIdentifier(
                                id=self.ldm_column_name, type="label"
                            )
                        ]
                    },
                    user={
                        "data": CatalogEntityIdentifier(id=udf_id, type="user")
                    },
                )
                user_data_filter = CatalogUserDataFilter(
                    id=udf_id,
                    attributes=attributes,
                    relationships=relationships,
                )

                try:
                    self._api.create_or_update_user_data_filter(
                        workspace_id, user_data_filter
                    )
                    self.logger.info(
                        "Created or updated user data filters for user with id "
                        + f"{udf_id} for client with id {workspace_id}"
                    )
                except Exception as e:
                    raise ContextException(
                        f"Failed to create user data filters: {e}",
                        udf_group,
                        user_data_filter,
                    ) from e

    def _delete_user_data_filters(
        self, workspace_id: str, udf_ids_to_delete: set[str]
    ) -> None:
        """Delete user data filters in GoodData workspaces."""
        for udf_id in udf_ids_to_delete:
            try:
                self._api.delete_user_data_filter(workspace_id, udf_id)
                self.logger.info(
                    f"Deleted user data filters for user with id {udf_id}"
                )
            except Exception as e:
                raise ContextException(
                    f"Failed to delete user data filters: {e}"
                ) from e

    def _provision_full_load(self) -> None:
        """Provision user data filters in GoodData workspaces."""

        if not self.maql_column_name:
            raise ContextException(
                "MAQL column name is not set. Please set it before provisioning."
            )
        if not self.ldm_column_name:
            raise ContextException(
                "LDM column name is not set. Please set it before provisioning."
            )

        grouped_db_user_data_filters = (
            self._group_db_user_data_filters_by_ws_id(self.source_group_full)
        )
        self._create_user_data_filters(grouped_db_user_data_filters)

    def _provision_incremental_load(self) -> None:
        """Provision user data filters in GoodData workspaces."""
        raise NotImplementedError("Not implemented yet.")



# (C) 2025 GoodData Corporation
from importlib import metadata

try:
    __version__ = metadata.version("gooddata-pipelines")
except metadata.PackageNotFoundError:
    __version__ = "unknown-version"



# (C) 2025 GoodData Corporation
import os
from pathlib import Path

import pytest
from pydantic import BaseModel, ValidationError

from gooddata_pipelines import (
    EntityType,
    PermissionFullLoad,
    PermissionIncrementalLoad,
    UserDataFilterFullLoad,
    UserFullLoad,
    UserGroupFullLoad,
    UserGroupIncrementalLoad,
    UserIncrementalLoad,
    WorkspaceFullLoad,
)
from gooddata_pipelines.provisioning.entities.workspaces.models import (
    WorkspaceIncrementalLoad,
)
from gooddata_pipelines.provisioning.provisioning import Provisioning
from tests.conftest import TEST_DATA_DIR

WORKSPACE_DATA_TO_FAIL = [
    WorkspaceFullLoad(
        parent_id="client_id1",
        workspace_id="workspace_id1",
        workspace_name="workspace_title1",
    ),
    WorkspaceIncrementalLoad(  # type: ignore
        parent_id="client_id1",
        workspace_id="workspace_id1",
        workspace_name="workspace_title1",
        is_active=True,
    ),
]
USER_DATA_TO_FAIL = [
    UserFullLoad(
        user_id="user_id1",
        firstname="user_name1",
        lastname="user_name1",
        email="user_email1",
        auth_id="auth_id1",
        user_groups=["user_group_id1"],
    ),
    UserIncrementalLoad(
        user_id="user_id1",
        firstname="user_name1",
        lastname="user_name1",
        email="user_email1",
        auth_id="auth_id1",
        user_groups=["user_group_id1"],
        is_active=True,
    ),
]

USER_GROUP_DATA_TO_FAIL = [
    UserGroupFullLoad(
        user_group_id="user_group_id1",
        user_group_name="user_group_name1",
    ),
    UserGroupIncrementalLoad(
        user_group_id="user_group_id1",
        user_group_name="user_group_name1",
        is_active=True,
    ),
]

PERMISSION_DATA_TO_FAIL = [
    PermissionFullLoad(
        permission="permission_id1",
        workspace_id="workspace_id1",
        entity_id="entity_id1",
        entity_type=EntityType.user,
    ),
    PermissionIncrementalLoad(
        permission="permission_id1",
        workspace_id="workspace_id1",
        entity_id="entity_id1",
        entity_type=EntityType.user,
        is_active=True,
    ),
]


TEST_CASES = [
    ("workspace_provisioner", WORKSPACE_DATA_TO_FAIL),
    ("user_provisioner", USER_DATA_TO_FAIL),
    ("user_group_provisioner", USER_GROUP_DATA_TO_FAIL),
    ("permission_provisioner", PERMISSION_DATA_TO_FAIL),
]


@pytest.mark.parametrize(
    "provisioner_name, data_to_fail",
    TEST_CASES,
)
def test_fail_type_validation(
    request: pytest.FixtureRequest, provisioner_name: str, data_to_fail: list
) -> None:
    """Data type validation of source data should fail when input data is not
    all of the same type."""
    provisioner = request.getfixturevalue(provisioner_name)
    with pytest.raises(TypeError) as e:
        provisioner._validate_source_data_type(
            data_to_fail,
            WorkspaceFullLoad,
        )

        assert "Not all elements in source data are instances of" in str(e)


def test_create_from_profile() -> None:
    """Test creating a provisioner from a profile."""

    os.environ["MOCK_TOKEN"] = "some_user_token"
    provisioner: Provisioning = Provisioning.create_from_profile(
        profile="mock_profile",
        profiles_path=Path(f"{TEST_DATA_DIR}/profiles.yaml"),
    )
    assert provisioner._api._domain == "http://localhost:3000"
    assert provisioner._api._token == os.environ.pop("MOCK_TOKEN")


MODELS_AND_VALID_DATA = [
    (
        WorkspaceFullLoad,
        {
            "parent_id": "parent_id",
            "workspace_id": "workspace_id",
            "workspace_name": "workspace_name",
        },
    ),
    (
        WorkspaceIncrementalLoad,
        {
            "parent_id": "parent_id",
            "workspace_id": "workspace_id",
            "workspace_name": "workspace_name",
            "is_active": True,
        },
    ),
    (
        UserFullLoad,
        {
            "user_id": "user_id",
            "firstname": "firstname",
            "lastname": "lastname",
            "email": "email",
            "auth_id": "auth_id",
            "user_groups": ["user_group_id"],
        },
    ),
    (
        UserIncrementalLoad,
        {
            "user_id": "user_id",
            "firstname": "firstname",
            "lastname": "lastname",
            "email": "email",
            "auth_id": "auth_id",
            "user_groups": ["user_group_id"],
            "is_active": True,
        },
    ),
    (
        UserGroupFullLoad,
        {
            "user_group_id": "user_group_id",
            "user_group_name": "user_group_name",
        },
    ),
    (
        UserGroupIncrementalLoad,
        {
            "user_group_id": "user_group_id",
            "user_group_name": "user_group_name",
            "is_active": True,
        },
    ),
    (
        PermissionFullLoad,
        {
            "permission": "permission",
            "workspace_id": "workspace_id",
            "entity_id": "entity_id",
            "entity_type": EntityType.user,
        },
    ),
    (
        PermissionIncrementalLoad,
        {
            "permission": "permission",
            "workspace_id": "workspace_id",
            "entity_id": "entity_id",
            "entity_type": EntityType.user,
            "is_active": True,
        },
    ),
    (
        UserDataFilterFullLoad,
        {
            "workspace_id": "workspace_id",
            "udf_id": "udf_id",
            "udf_value": "udf_value",
        },
    ),
]


@pytest.mark.parametrize(["provisioning_model", "data"], MODELS_AND_VALID_DATA)
def test_validate_valid_data(provisioning_model: BaseModel, data: dict) -> None:
    """Test validating valid data."""
    provisioning_model.model_validate(data)


@pytest.mark.parametrize(
    ["provisioning_model", "data"],
    MODELS_AND_VALID_DATA,
)
def test_raise_extra_fields(provisioning_model: BaseModel, data: dict) -> None:
    """Test raising an error when extra fields are provided."""
    data["extra_field"] = "extra_field"
    with pytest.raises(ValidationError):
        provisioning_model.model_validate(data)



# (C) 2025 GoodData Corporation

from enum import Enum
from typing import Iterator, TypeAlias

import attrs
from gooddata_sdk.catalog.identifier import CatalogAssigneeIdentifier
from gooddata_sdk.catalog.permission.declarative_model.permission import (
    CatalogDeclarativeSingleWorkspacePermission,
    CatalogDeclarativeWorkspacePermissions,
)
from pydantic import BaseModel, ConfigDict

from gooddata_pipelines.provisioning.utils.exceptions import BaseUserException

TargetsPermissionDict: TypeAlias = dict[str, dict[str, bool]]


class EntityType(str, Enum):
    # NOTE: Start using StrEnum with Python 3.11
    user = "user"
    user_group = "userGroup"


class BasePermission(BaseModel):
    model_config = ConfigDict(extra="forbid")

    permission: str
    workspace_id: str
    entity_id: str
    entity_type: EntityType


class PermissionFullLoad(BasePermission):
    """Input validator for full load of workspace permissions provisioning."""


class PermissionIncrementalLoad(BasePermission):
    """Input validator for incremental load of workspace permissions provisioning."""

    is_active: bool


@attrs.define
class PermissionDeclaration:
    users: TargetsPermissionDict
    user_groups: TargetsPermissionDict

    @classmethod
    def from_sdk_api(
        cls, declaration: CatalogDeclarativeWorkspacePermissions
    ) -> "PermissionDeclaration":
        """
        Constructs an WSPermissionDeclaration instance
        from GoodData SDK CatalogDeclarativeWorkspacePermissions.
        """
        users: TargetsPermissionDict = {}
        user_groups: TargetsPermissionDict = {}

        for permission in declaration.permissions:
            permission_type, id = (
                permission.assignee.type,
                permission.assignee.id,
            )

            if permission_type == EntityType.user.value:
                target_dict = users
            else:
                target_dict = user_groups

            id_permissions = target_dict.get(id)
            if not id_permissions:
                target_dict[id] = dict()

            target_dict[id][permission.name] = True

        return PermissionDeclaration(users, user_groups)

    @staticmethod
    def _construct_upstream_permission(
        permission: str, assignee: CatalogAssigneeIdentifier
    ) -> CatalogDeclarativeSingleWorkspacePermission | None:
        """Constructs single permission declaration for the SDK API."""
        try:
            return CatalogDeclarativeSingleWorkspacePermission(
                name=permission, assignee=assignee
            )
        except Exception as e:
            raise BaseUserException(
                f"Failed to construct SDK declaration for type={assignee.type} ",
                f"id={assignee.id}. Error: {e}",
            )

    def _permissions_for_target(
        self, permissions: dict[str, bool], assignee: CatalogAssigneeIdentifier
    ) -> Iterator[CatalogDeclarativeSingleWorkspacePermission]:
        """Constructs permission declarations for a single target."""
        for permission, is_active in permissions.items():
            if not is_active:
                continue
            declaration = self._construct_upstream_permission(
                permission, assignee
            )
            if not declaration:
                continue
            yield declaration

    def to_sdk_api(self) -> CatalogDeclarativeWorkspacePermissions:
        """
        Constructs the GoodData SDK CatalogDeclarativeWorkspacePermissions
        object from the WSPermissionDeclaration instance.
        """
        permission_declarations: list[
            CatalogDeclarativeSingleWorkspacePermission
        ] = []

        for user_id, permissions in self.users.items():
            assignee = CatalogAssigneeIdentifier(
                id=user_id, type=EntityType.user.value
            )
            for declaration in self._permissions_for_target(
                permissions, assignee
            ):
                permission_declarations.append(declaration)

        for ug_id, permissions in self.user_groups.items():
            assignee = CatalogAssigneeIdentifier(
                id=ug_id, type=EntityType.user_group.value
            )
            for declaration in self._permissions_for_target(
                permissions, assignee
            ):
                permission_declarations.append(declaration)

        return CatalogDeclarativeWorkspacePermissions(
            permissions=permission_declarations
        )

    def add_incremental_permission(
        self, permission: PermissionIncrementalLoad
    ) -> None:
        """
        Adds WSPermission object into respective field within the instance.
        Handles duplicate permissions and different combinations of input
        and upstream is_active permission states.
        """
        target_dict = (
            self.users
            if permission.entity_type == EntityType.user
            else self.user_groups
        )

        if permission.entity_id not in target_dict:
            target_dict[permission.entity_id] = {}

        is_active = permission.is_active
        target_permissions = target_dict[permission.entity_id]
        permission_value = permission.permission

        if permission_value not in target_permissions:
            target_permissions[permission_value] = is_active
        elif not is_active and target_permissions[permission_value] is True:
            print(
                "isActive=False provided after True has been specificed for the "
                + f"same input. Skipping '{permission}'"
            )
        elif is_active and target_permissions[permission_value] is False:
            print(
                "isActive=True provided after False has been specified for the "
                + f"same input. Overwriting '{permission}'"
            )
            target_permissions[permission_value] = is_active

    def add_full_load_permission(self, permission: PermissionFullLoad) -> None:
        """
        Adds WSPermission object into respective field within the instance.
        Handles duplicate permissions and different combinations of input
        and upstream is_active permission states.
        """
        target_dict = (
            self.users
            if permission.entity_type == EntityType.user
            else self.user_groups
        )

        if permission.entity_id not in target_dict:
            target_dict[permission.entity_id] = {}

        target_permissions = target_dict[permission.entity_id]
        permission_value = permission.permission

        if permission_value not in target_permissions:
            target_permissions[permission_value] = True

    def upsert(self, other: "PermissionDeclaration") -> None:
        """
        Modifies the owner object by merging with the other.
        Keeps the unmodified users/userGroups untouched.
        If some user/userGroup is modified, it gets overwritten with permissions
        defined in the input.
        """
        for user_id, permissions in other.users.items():
            self.users[user_id] = permissions

        for ug_id, permissions in other.user_groups.items():
            self.user_groups[ug_id] = permissions


WSPermissionsDeclarations: TypeAlias = dict[str, PermissionDeclaration]



# (C) 2025 GoodData Corporation

import json
from pathlib import Path
from typing import Any

import attrs
import yaml


class PathUtils:
    """Handles common path operations."""

    @staticmethod
    def validate_path(path: str | Path) -> Path:
        """Validates a path."""
        if not isinstance(path, Path):
            path = Path(path)

        return path

    def check_path_exists(self, path: Path) -> None:
        """Checks if a path exists."""
        if not self.validate_path(path).exists():
            raise FileNotFoundError(f"File {path} does not exist.")


@attrs.define
class JsonUtils:
    """Handles common JSON interactions."""

    path_utils: PathUtils = attrs.field(factory=PathUtils)

    def load(self, path: Path) -> Any:
        """Loads a JSON file."""
        self.path_utils.check_path_exists(path)

        with open(path, "r") as f:
            return json.load(f)

    def dump(self, path: Path, data: Any) -> None:
        """Writes the source to a JSON file."""
        with open(path, "w") as output_file:
            json.dump(data, output_file)


@attrs.define
class YamlUtils:
    """Handles common YMAL interactions."""

    path_utils: PathUtils = attrs.field(factory=PathUtils)

    def safe_load(self, path: Path) -> Any:
        """Safe loads a YAML file."""
        self.path_utils.check_path_exists(path)

        with open(path, "r") as f:
            return yaml.safe_load(f)

    def dump(self, path: str, data: Any) -> None:
        """Writes the source to a YAML file."""
        with open(path, "w") as output_file:
            yaml.dump(data, output_file)



# (C) 2025 GoodData Corporation
import pytest

from gooddata_pipelines.ldm_extension.input_processor import (
    LdmExtensionDataProcessor,
)
from gooddata_pipelines.ldm_extension.models.custom_data_object import (
    ColumnDataType,
    CustomDataset,
    CustomDatasetDefinition,
    CustomFieldDefinition,
    CustomFieldType,
)


@pytest.fixture
def mock_custom_field_attribute():
    return CustomFieldDefinition(
        workspace_id="workspace1",
        dataset_id="ds1",
        custom_field_id="attr1",
        custom_field_name="Attribute 1",
        custom_field_type=CustomFieldType.ATTRIBUTE,
        custom_field_source_column="col_attr1",
        custom_field_source_column_data_type=ColumnDataType.STRING,
    )


@pytest.fixture
def mock_custom_field_fact():
    return CustomFieldDefinition(
        workspace_id="workspace1",
        dataset_id="ds1",
        custom_field_id="fact1",
        custom_field_name="Fact 1",
        custom_field_type=CustomFieldType.FACT,
        custom_field_source_column="col_fact1",
        custom_field_source_column_data_type=ColumnDataType.INT,
    )


@pytest.fixture
def mock_custom_field_date():
    return CustomFieldDefinition(
        workspace_id="workspace1",
        dataset_id="ds1",
        custom_field_id="date1",
        custom_field_name="Date 1",
        custom_field_type=CustomFieldType.DATE,
        custom_field_source_column="col_date1",
        custom_field_source_column_data_type=ColumnDataType.DATE,
    )


@pytest.fixture
def mock_dataset_definition():
    return CustomDatasetDefinition(
        workspace_id="workspace1",
        dataset_id="ds1",
        dataset_name="Dataset 1",
        dataset_source_table="table1",
        dataset_datasource_id="ds_source",
        dataset_source_sql=None,
        parent_dataset_reference="parent_ds",
        parent_dataset_reference_attribute_id="parent_attr",
        dataset_reference_source_column="ref_col",
        dataset_reference_source_column_data_type=ColumnDataType.STRING,
        workspace_data_filter_id="wdf1",
        workspace_data_filter_column_name="col1",
    )


@pytest.fixture
def mock_custom_dataset(
    mock_dataset_definition,
    mock_custom_field_attribute,
    mock_custom_field_fact,
    mock_custom_field_date,
):
    return CustomDataset(
        definition=mock_dataset_definition,
        custom_fields=[
            mock_custom_field_attribute,
            mock_custom_field_fact,
            mock_custom_field_date,
        ],
    )


def test_attribute_from_field(mock_custom_field_attribute):
    attr = LdmExtensionDataProcessor._attribute_from_field(
        "dataset_name", mock_custom_field_attribute
    )
    assert attr.id == "attr1"
    assert attr.title == "Attribute 1"
    assert attr.source_column == "col_attr1"
    assert attr.source_column_data_type == ColumnDataType.STRING.value
    assert attr.tags == ["dataset_name"]


def test_fact_from_field(mock_custom_field_fact):
    fact = LdmExtensionDataProcessor._fact_from_field(
        "dataset_name", mock_custom_field_fact
    )
    assert fact.id == "fact1"
    assert fact.title == "Fact 1"
    assert fact.source_column == "col_fact1"
    assert fact.source_column_data_type == ColumnDataType.INT.value
    assert fact.tags == ["dataset_name"]


def test_date_from_field(mock_custom_field_date):
    processor = LdmExtensionDataProcessor()
    date_ds = processor._date_from_field("dataset_name", mock_custom_field_date)
    assert date_ds.id == "date1"
    assert date_ds.title == "Date 1"
    assert set(date_ds.granularities) == set(processor.DATE_GRANULARITIES)
    assert date_ds.tags == ["dataset_name"]


def test_date_ref_from_field(mock_custom_field_date):
    ref = LdmExtensionDataProcessor._date_ref_from_field(mock_custom_field_date)
    assert ref.identifier.id == "date1"
    assert ref.sources
    assert ref.sources[0].column == "col_date1"
    assert ref.sources[0].data_type == ColumnDataType.DATE.value


def test_get_sources_table_only(mock_dataset_definition):
    mock_dataset_definition.dataset_source_sql = None
    dataset = CustomDataset(
        definition=mock_dataset_definition, custom_fields=[]
    )
    table_id, sql = LdmExtensionDataProcessor._get_sources(dataset)
    assert table_id is not None
    assert table_id.id == "table1"
    assert sql is None


def test_get_sources_sql_only(mock_dataset_definition):
    mock_dataset_definition.dataset_source_table = None
    mock_dataset_definition.dataset_source_sql = "SELECT * FROM foo"
    dataset = CustomDataset(
        definition=mock_dataset_definition, custom_fields=[]
    )
    table_id, sql = LdmExtensionDataProcessor._get_sources(dataset)
    assert table_id is None
    assert sql is not None
    assert sql.statement == "SELECT * FROM foo"


def test_datasets_to_ldm(mock_custom_dataset):
    print(mock_custom_dataset)
    processor = LdmExtensionDataProcessor()
    datasets = {"ds1": mock_custom_dataset}
    model = processor.datasets_to_ldm(datasets)
    # Check that the model contains the expected dataset and date instance
    ldm = model.ldm
    assert ldm
    assert len(ldm.datasets) == 1
    ds = ldm.datasets[0]
    assert ds.id == "ds1"
    assert ds.title == "Dataset 1"
    assert ds.attributes
    assert ds.facts
    assert len(ds.attributes) == 1
    assert len(ds.facts) == 1
    assert len(ds.references) == 2  # 1 parent + 1 date
    assert ds.workspace_data_filter_columns
    assert ds.workspace_data_filter_references
    assert ds.workspace_data_filter_columns[0].name == "col1"
    assert ds.workspace_data_filter_references[0].filter_id.id == "wdf1"
    assert len(ldm.date_instances) == 1
    assert ldm.date_instances[0].id == "date1"



# (C) 2025 GoodData Corporation

import os
import shutil
import tempfile
import time
import traceback
from pathlib import Path
from typing import Any

import attrs
import requests

from gooddata_pipelines.backup_and_restore.backup_input_processor import (
    BackupInputProcessor,
)
from gooddata_pipelines.backup_and_restore.base_manager import BaseManager
from gooddata_pipelines.backup_and_restore.constants import (
    BackupSettings,
    DirNames,
)
from gooddata_pipelines.backup_and_restore.models.input_type import InputType
from gooddata_pipelines.backup_and_restore.models.storage import (
    BackupRestoreConfig,
)
from gooddata_pipelines.backup_and_restore.storage.base_storage import (
    BackupStorage,
)
from gooddata_pipelines.utils.rate_limiter import RateLimiter


@attrs.define
class BackupBatch:
    list_of_ids: list[str]


class BackupManager(BaseManager):
    storage: BackupStorage

    def __init__(self, host: str, token: str, config: BackupRestoreConfig):
        super().__init__(host, token, config)

        self.org_id = self._api.get_organization_id()

        self.loader = BackupInputProcessor(self._api, self.config.api_page_size)

        self._api_rate_limiter = RateLimiter(
            calls_per_second=self.config.api_calls_per_second,
        )

    def get_user_data_filters(self, ws_id: str) -> dict:
        """Returns the user data filters for the specified workspace."""
        with self._api_rate_limiter:
            response: requests.Response = self._api.get_user_data_filters(ws_id)
            if response.ok:
                return response.json()
            else:
                raise RuntimeError(f"{response.status_code}: {response.text}")

    def _store_user_data_filters(
        self,
        user_data_filters: dict,
        export_path: Path,
        ws_id: str,
    ) -> None:
        """Stores the user data filters in the specified export path."""
        os.mkdir(
            os.path.join(
                export_path,
                "gooddata_layouts",
                self.org_id,
                "workspaces",
                ws_id,
                "user_data_filters",
            )
        )

        for filter in user_data_filters["userDataFilters"]:
            udf_file_path = os.path.join(
                export_path,
                "gooddata_layouts",
                self.org_id,
                "workspaces",
                ws_id,
                "user_data_filters",
                filter["id"] + ".yaml",
            )
            self.yaml_utils.dump(udf_file_path, filter)

    @staticmethod
    def _move_folder(source: Path, destination: Path) -> None:
        """Moves the source folder to the destination."""
        shutil.move(source, destination)

    def _get_automations_from_api(self, workspace_id: str) -> Any:
        """Returns automations for the workspace as JSON."""
        with self._api_rate_limiter:
            response: requests.Response = self._api.get_automations(
                workspace_id
            )
            if response.ok:
                return response.json()
            else:
                raise RuntimeError(
                    f"Failed to get automations for {workspace_id}. "
                    + f"{response.status_code}: {response.text}"
                )

    def _store_automations(self, export_path: Path, workspace_id: str) -> None:
        """Stores the automations in the specified export path."""
        # Get the automations from the API
        automations: Any = self._get_automations_from_api(workspace_id)

        automations_folder_path: Path = Path(
            export_path,
            "gooddata_layouts",
            self.org_id,
            "workspaces",
            workspace_id,
            "automations",
        )

        automations_file_path: Path = Path(
            automations_folder_path, "automations.json"
        )

        os.mkdir(automations_folder_path)

        # Store the automations in a JSON file
        if len(automations["data"]) > 0:
            self.json_utils.dump(automations_file_path, automations)

    def store_declarative_filter_views(
        self, export_path: Path, workspace_id: str
    ) -> None:
        """Stores the filter views in the specified export path."""
        # Get the filter views YAML files from the API
        with self._api_rate_limiter:
            self._api.store_declarative_filter_views(workspace_id, export_path)

        # Move filter views to the subfolder containing the analytics model
        self._move_folder(
            Path(export_path, "gooddata_layouts", self.org_id, "filter_views"),
            Path(
                export_path,
                "gooddata_layouts",
                self.org_id,
                "workspaces",
                workspace_id,
                "filter_views",
            ),
        )

    def _get_workspace_export(
        self,
        local_target_path: str,
        workspaces_to_export: list[str],
    ) -> None:
        """
        Iterate over all workspaces in the workspaces_to_export list and store
        their declarative_workspace and their respective user data filters.
        """
        exported = False
        for workspace_id in workspaces_to_export:
            export_path = Path(
                local_target_path,
                self.org_id,
                workspace_id,
                BackupSettings.TIMESTAMP_SDK_FOLDER,
            )

            try:
                user_data_filters = self.get_user_data_filters(workspace_id)
            except Exception as e:
                self.logger.error(
                    f"Skipping backup of {workspace_id} - check if workspace exists."
                    + f"{e.__class__.__name__}: {e}"
                )
                continue

            try:
                # TODO: consider using the API to get JSON declarations in memory
                # or check if there is a way to get YAML structures directly from
                # the SDK. That way we could save and package all the declarations
                # directly instead of reorganizing the folder structures. That should
                # be more transparent/readable and possibly safer for threading
                with self._api_rate_limiter:
                    self._api.store_declarative_workspace(
                        workspace_id, export_path
                    )
                self.store_declarative_filter_views(export_path, workspace_id)
                self._store_automations(export_path, workspace_id)

                self._store_user_data_filters(
                    user_data_filters, export_path, workspace_id
                )
                self.logger.info(f"Stored export for {workspace_id}")
                exported = True
            except Exception as e:
                self.logger.error(
                    f"Skipping {workspace_id}. {e.__class__.__name__} encountered: {e}"
                )

        if not exported:
            raise RuntimeError(
                "None of the workspaces were exported. Check that the source file "
                + "is correct and that the workspaces exist."
            )

    def _archive_gooddata_layouts_to_zip(self, folder: str) -> None:
        """Archives the gooddata_layouts directory to a zip file."""
        try:
            target_subdir = ""
            for subdir, dirs, files in os.walk(folder):
                if DirNames.LAYOUTS in dirs:
                    target_subdir = os.path.join(subdir, dirs[0])
                if DirNames.LDM in dirs:
                    inner_layouts_dir = subdir + "/gooddata_layouts"
                    os.mkdir(inner_layouts_dir)
                    for dir in dirs:
                        shutil.move(
                            os.path.join(subdir, dir),
                            os.path.join(inner_layouts_dir),
                        )
                    shutil.make_archive(target_subdir, "zip", subdir)
                    shutil.rmtree(target_subdir)
        except Exception as e:
            self.logger.error(f"Error archiving {folder} to zip: {e}")
            raise

    @staticmethod
    def _split_to_batches(
        workspaces_to_export: list[str], batch_size: int
    ) -> list[BackupBatch]:
        """Splits the list of workspaces into batches of the specified size.
        The batch is represented as a list of workspace IDs.
        Returns a list of batches (i.e. list of lists of IDs)
        """
        list_of_batches = []
        while workspaces_to_export:
            batch = BackupBatch(workspaces_to_export[:batch_size])
            workspaces_to_export = workspaces_to_export[batch_size:]
            list_of_batches.append(batch)

        return list_of_batches

    def _process_batch(
        self,
        batch: BackupBatch,
        retry_count: int = 0,
    ) -> None:
        """Processes a single batch of workspaces for backup.
        If the batch processing fails, the function will wait
        and retry with exponential backoff up to BackupSettings.MAX_RETRIES.
        The base wait time is defined by BackupSettings.RETRY_DELAY.
        """
        try:
            with tempfile.TemporaryDirectory() as tmpdir:
                self._get_workspace_export(tmpdir, batch.list_of_ids)

                self._archive_gooddata_layouts_to_zip(
                    str(Path(tmpdir, self.org_id))
                )

                self.storage.export(tmpdir, self.org_id)

        except Exception as e:
            if retry_count < BackupSettings.MAX_RETRIES:
                # Retry with exponential backoff until MAX_RETRIES
                next_retry = retry_count + 1
                wait_time = BackupSettings.RETRY_DELAY**next_retry
                self.logger.info(
                    f"{e.__class__.__name__} encountered while processing a batch. "
                    + f"Retrying {next_retry}/{BackupSettings.MAX_RETRIES} "
                    + f"in {wait_time} seconds..."
                )

                time.sleep(wait_time)
                self._process_batch(batch, next_retry)
            else:
                # If the batch fails after MAX_RETRIES, raise the error
                self.logger.error(f"Batch failed: {e.__class__.__name__}: {e}")
                raise

    def _process_batches(
        self,
        batches: list[BackupBatch],
    ) -> None:
        """
        Processes batches sequentially to avoid overloading the API.
        If any batch fails, the processing will stop.
        """
        for i, batch in enumerate(batches, 1):
            self.logger.info(f"Processing batch {i}/{len(batches)}...")
            self._process_batch(batch)

    def backup_workspaces(
        self,
        path_to_csv: str | None = None,
        workspace_ids: list[str] | None = None,
    ) -> None:
        """Runs the backup process for a list of workspace IDs.

        Will take the list of workspace IDs or read the list of
        workspace IDs from a CSV file and create backup for each
        workspace in storage specified in the configuration.

        Args:
            path_to_csv (str): Path to a CSV file containing a list of workspace IDs
            workspace_ids (list[str]): List of workspace IDs
        """
        self._backup(InputType.LIST_OF_WORKSPACES, path_to_csv, workspace_ids)

    def backup_hierarchies(
        self,
        path_to_csv: str | None = None,
        workspace_ids: list[str] | None = None,
    ) -> None:
        """Runs the backup process for a list of hierarchies.

        Will take the list of workspace IDs or read the list of workspace IDs
        from a CSV file and create backup for each of those workspaces' hierarchies
        in storage specified in the configuration.
        Workspace hierarchy means the workspace itself and all its direct and
        indirect children.

        Args:
            path_to_csv (str): Path to a CSV file containing a list of workspace IDs
            workspace_ids (list[str]): List of workspace IDs
        """
        self._backup(InputType.HIERARCHY, path_to_csv, workspace_ids)

    def backup_entire_organization(self) -> None:
        """Runs the backup process for the entire organization.

        Will create backup for all workspaces in the organization in storage
        specified in the configuration.
        """
        self._backup(InputType.ORGANIZATION)

    def _backup(
        self,
        input_type: InputType,
        path_to_csv: str | None = None,
        workspace_ids: list[str] | None = None,
    ) -> None:
        """Runs the backup process with the selected input type."""
        try:
            workspaces_to_export: list[str] = self.loader.get_ids_to_backup(
                input_type,
                path_to_csv,
                workspace_ids,
            )
            batches = self._split_to_batches(
                workspaces_to_export, self.config.batch_size
            )

            self.logger.info(
                f"Exporting {len(workspaces_to_export)} workspaces in {len(batches)} batches."
            )

            self._process_batches(batches)

            self.logger.info("Backup completed")
        except Exception as e:
            self.logger.error(f"Backup failed: {e.__class__.__name__}: {e}")
            self.logger.error(traceback.format_exc())
            raise



# (C) 2025 GoodData Corporation
"""Module for validating custom fields input data.

This module is responsible for validating custom fields input data checking for
row level and aggregated constraints.
"""

from collections import Counter
from typing import Any, TypeVar

from pydantic import BaseModel

from gooddata_pipelines.ldm_extension.models.aliases import (
    DatasetId,
    WorkspaceId,
)
from gooddata_pipelines.ldm_extension.models.custom_data_object import (
    CustomDataset,
    CustomDatasetDefinition,
    CustomFieldDefinition,
    CustomFieldType,
)


class LdmExtensionDataValidator:
    ModelT = TypeVar("ModelT", bound=BaseModel)

    def validate(
        self,
        dataset_definitions: list[CustomDatasetDefinition],
        field_definitions: list[CustomFieldDefinition],
    ) -> dict[WorkspaceId, dict[DatasetId, CustomDataset]]:
        """Validate dataset and field definitions.

        Validates the dataset definitions and field definitions by using Pydantic
        models to check row level constraints, then aggregates the definitions
        per workspace, while checking for integrity on aggregated level, i.e.,
        uniqueness of combinations of identifieres on workspace level.

        Args:
            raw_dataset_definitions (list[dict[str, str]]): List of raw dataset definitions to validate.
            raw_field_definitions (list[dict[str, str]]): List of raw field definitions to validate.
        Returns:
            dict[WorkspaceId, dict[DatasetId, CustomDataset]]:
                Dictionary of validated dataset definitions per workspace,
                where each dataset contains its custom fields:
                ```python
                {
                    "workspace_id_1": {
                        "dataset_id_1": CustomDataset(...),
                        "dataset_id_2": CustomDataset(...),
                    },
                    ...
                }
                ```
        """

        # First, validate the dataset definitions and aggregate them per workspace.
        validated_data = self._validate_dataset_definitions(dataset_definitions)

        # Then validate the field definitions and connect them to the datasets
        validated_data = self._validate_field_definitions(
            validated_data, field_definitions
        )

        return validated_data

    def _validate_dataset_definitions(
        self,
        dataset_definitions: list[CustomDatasetDefinition],
    ) -> dict[WorkspaceId, dict[DatasetId, CustomDataset]]:
        self._check_dataset_combinations(dataset_definitions)

        validated_definitions: dict[
            WorkspaceId, dict[DatasetId, CustomDataset]
        ] = {}
        for definition in dataset_definitions:
            validated_definitions.setdefault(definition.workspace_id, {})[
                definition.dataset_id
            ] = CustomDataset(definition=definition, custom_fields=[])

        return validated_definitions

    def _check_dataset_combinations(
        self, dataset_definitions: list[CustomDatasetDefinition]
    ) -> None:
        """Check integrity of provided dataset definitions.

        Validation criteria:
            - workspace_id + dataset_id must be unique across all dataset definitions.

        Args:
            dataset_definitions (list[CustomDatasetDefinition]): List of dataset definitions to check.
        Raises:
            ValueError: If there are duplicate dataset definitions based on workspace_id and dataset_id.
        """
        workspace_dataset_combinations = [
            (definition.workspace_id, definition.dataset_id)
            for definition in dataset_definitions
        ]
        if len(workspace_dataset_combinations) != len(
            set(workspace_dataset_combinations)
        ):
            duplicates = self._get_duplicates(workspace_dataset_combinations)
            raise ValueError(
                "Duplicate dataset definitions found in the raw dataset "
                + f"definitions (workspace_id, dataset_id): {duplicates}"
            )

    @staticmethod
    def _get_duplicates(list_to_check: list[Any]) -> list[Any]:
        """Get duplicates from a list.

        Args:
            list_to_check (list[Any]): List of items to check for duplicates.
        Returns:
            list[Any]: List of duplicate items.
        """
        counts = Counter(list_to_check)
        return [item for item, count in counts.items() if count > 1]

    def _check_field_combinations(
        self, field_definitions: list[CustomFieldDefinition]
    ) -> None:
        """Check integrity of provided field definitions.

        Validation criteria (per workspace):
            - unique workspace_id + cf_id combinations (only for attribute and fact custom_field_type)
            - there is no row with the same dataset_id and cf_id (only for date custom_field_type)

        Args:
            field_definitions (list[CustomFieldDefinition]): List of field definitions to check.
        Raises:
            ValueError: If there are duplicate field definitions based on workspace_id and cf_id.
        """
        workspace_field_combinations: set[tuple[str, str]] = set()
        dataset_field_combinations: set[tuple[str, str]] = set()

        for field in field_definitions:
            if field.custom_field_type in [
                CustomFieldType.ATTRIBUTE,
                CustomFieldType.FACT,
            ]:
                combination = (field.workspace_id, field.custom_field_id)
                if combination in workspace_field_combinations:
                    raise ValueError(
                        f"Duplicate custom field found for workspace {field.workspace_id} "
                        + f"with field ID {field.custom_field_id}"
                    )
                workspace_field_combinations.add(combination)

            elif field.custom_field_type == CustomFieldType.DATE:
                combination = (field.dataset_id, field.custom_field_id)
                if combination in dataset_field_combinations:
                    raise ValueError(
                        f"Duplicate custom field found for dataset {field.dataset_id} "
                        + f"with field ID {field.custom_field_id}"
                    )
                dataset_field_combinations.add(combination)

    def _validate_field_definitions(
        self,
        validated_definitions: dict[
            WorkspaceId, dict[DatasetId, CustomDataset]
        ],
        field_definitions: list[CustomFieldDefinition],
    ) -> dict[WorkspaceId, dict[DatasetId, CustomDataset]]:
        """Validates custom field definitions amd connects them to the datasets.

        Args:
            validated_definitions (dict[WorkspaceId, dict[DatasetId, CustomDataset]]):
                Dictionary of validated dataset definitions per workspace.
            raw_field_definitions (list[dict[str, str]]): List of raw field definitions to validate.
        Returns:
            dict[WorkspaceId, dict[DatasetId, CustomDataset]]:
                Updated dictionary of validated dataset definitions with custom fields added.
        """
        self._check_field_combinations(field_definitions)

        for field_definition in field_definitions:
            validated_definitions[field_definition.workspace_id][
                field_definition.dataset_id
            ].custom_fields.append(field_definition)

        return validated_definitions



# (C) 2025 GoodData Corporation
import pytest
from pytest_mock import MockerFixture

from gooddata_pipelines.ldm_extension.ldm_extension_manager import (
    LdmExtensionManager,
)
from gooddata_pipelines.ldm_extension.models.analytical_object import (
    AnalyticalObject,
    Attributes,
)


@pytest.fixture
def manager(mocker: MockerFixture, mock_logger):
    custom_fields_manager = LdmExtensionManager(host="host", token="token")

    mocker.patch.object(custom_fields_manager, "_validator")
    mocker.patch.object(custom_fields_manager, "_processor")
    mocker.patch.object(custom_fields_manager, "_sdk")
    mocker.patch.object(custom_fields_manager, "_api")

    custom_fields_manager.logger.subscribe(mock_logger)

    return custom_fields_manager


@pytest.fixture
def validated_data(mocker: MockerFixture):
    # Minimal valid structure for validated_data
    return {"workspace_1": {"dataset_1": mocker.MagicMock()}}


def make_analytical_object(
    id, title="Title", type="type", are_relations_valid=True
):
    obj = AnalyticalObject(
        id=id,
        type=type,
        attributes=Attributes(
            title=title, areRelationsValid=are_relations_valid
        ),
    )
    return obj


def test_relations_check_success(
    manager, validated_data, mocker: MockerFixture
):
    """Relation check passes, workspace layout not reverted."""
    # Setup mocks
    mocker.patch.object(
        manager._sdk.catalog_workspace,
        "get_declarative_workspace",
        return_value=mocker.MagicMock(
            json=mocker.MagicMock(return_value="layout_json")
        ),
    )
    mocker.patch.object(
        manager,
        "_get_analytical_objects",
        side_effect=[
            [make_analytical_object("a", "A")],  # current
            [make_analytical_object("a", "A")],  # new
        ],
    )
    mocker.patch.object(
        manager,
        "_get_objects_with_invalid_relations",
        side_effect=[
            set(),  # current_invalid_relations
            set(),  # new_invalid_relations
        ],
    )
    mocker.patch.object(
        manager._processor, "datasets_to_ldm", return_value="ldm"
    )
    mocker.patch.object(
        manager._sdk.catalog_workspace_content, "put_declarative_ldm"
    )
    mocker.patch.object(
        manager, "_new_ldm_does_not_invalidate_relations", return_value=True
    )
    mocker.patch.object(
        manager._sdk.catalog_workspace, "put_declarative_workspace"
    )

    # Should print "Workspace workspace_1 LDM updated." and not revert
    manager._process_with_relations_check(validated_data)
    manager._sdk.catalog_workspace_content.put_declarative_ldm.assert_called_once()
    manager._sdk.catalog_workspace.put_declarative_workspace.assert_not_called()


def test_relations_check_failure_and_revert(
    manager, validated_data, capsys, mocker: MockerFixture
):
    """Relation check fails, workspace layout is reverted."""
    # Setup mocks
    mocker.patch.object(manager._api, "get_workspace_layout")
    obj1 = make_analytical_object("a", "A", "type", False)
    obj2 = make_analytical_object("b", "B", "type", False)
    mocker.patch.object(
        manager,
        "_get_objects_with_invalid_relations",
        side_effect=[
            [obj1],  # current_invalid_relations
            [obj1, obj2],  # new_invalid_relations (one more invalid)
        ],
    )
    mocker.patch.object(
        manager._processor, "datasets_to_ldm", return_value="ldm"
    )
    mocker.patch.object(
        manager._sdk.catalog_workspace_content, "put_declarative_ldm"
    )
    mocker.patch.object(
        manager, "_new_ldm_does_not_invalidate_relations", return_value=False
    )
    mocker.patch.object(
        manager._sdk.catalog_workspace, "put_declarative_workspace"
    )

    manager._process_with_relations_check(validated_data)

    # Should revert and print info about invalid relations
    manager._sdk.catalog_workspace.put_declarative_workspace.assert_called_once()
    out = capsys.readouterr().out
    assert (
        "Difference in invalid relations found in workspace workspace_1." in out
    )
    assert "b (type) B" in out
    assert "Reverting the workspace layout to the original state." in out


def test_relations_check_fewer_invalid_relations(
    manager, validated_data, mocker: MockerFixture
):
    """Fewer invalid relations after LDM update, no revert needed."""
    # Setup mocks
    obj1 = make_analytical_object("a", "A", "type", False)
    mocker.patch.object(
        manager._sdk.catalog_workspace,
        "get_declarative_workspace",
        return_value=mocker.MagicMock(
            json=mocker.MagicMock(return_value="layout_json")
        ),
    )
    mocker.patch.object(
        manager,
        "_get_objects_with_invalid_relations",
        side_effect=[
            [
                obj1,
                make_analytical_object("b", "B", "type", False),
            ],  # current_invalid_relations
            [obj1],  # new_invalid_relations (fewer)
        ],
    )
    mocker.patch.object(
        manager._processor, "datasets_to_ldm", return_value="ldm"
    )
    mocker.patch.object(
        manager._sdk.catalog_workspace_content, "put_declarative_ldm"
    )
    mocker.patch.object(
        manager, "_new_ldm_does_not_invalidate_relations", return_value=True
    )
    mocker.patch.object(
        manager._sdk.catalog_workspace, "put_declarative_workspace"
    )

    manager._process_with_relations_check(validated_data)
    manager._sdk.catalog_workspace.put_declarative_workspace.assert_not_called()


def test_log_diff_invalid_relations(manager, capsys):
    """Log diff invalid relations."""
    manager._log_diff_invalid_relations(
        [
            make_analytical_object("a", "A", "type", False),
            make_analytical_object("c", "C", "type", False),
        ],
        [
            make_analytical_object("b", "B", "type", False),
            make_analytical_object("c", "C", "type", False),
            make_analytical_object("d", "D", "type", False),
        ],
    )

    captured_output = capsys.readouterr().out

    assert "b (type) B" in captured_output
    assert "d (type) D" in captured_output
    assert "c (type) C" not in captured_output



# (C) 2025 GoodData Corporation
from typing import Literal

import orjson
import pytest
from gooddata_api_client.exceptions import (  # type: ignore[import]
    NotFoundException,
)
from gooddata_sdk.catalog.identifier import CatalogAssigneeIdentifier
from gooddata_sdk.catalog.permission.declarative_model.permission import (
    CatalogDeclarativeSingleWorkspacePermission,
    CatalogDeclarativeWorkspacePermissions,
)
from pytest_mock import MockerFixture

from gooddata_pipelines.provisioning.entities.users.models.permissions import (
    EntityType,
    PermissionDeclaration,
    PermissionFullLoad,
    PermissionIncrementalLoad,
)
from gooddata_pipelines.provisioning.entities.users.permissions import (
    PermissionProvisioner,
)
from tests.conftest import TEST_DATA_DIR

TEST_DATA_SUBDIR = f"{TEST_DATA_DIR}/provisioning/entities/permissions"

USER_1 = CatalogAssigneeIdentifier(id="user_1", type="user")
USER_2 = CatalogAssigneeIdentifier(id="user_2", type="user")
USER_3 = CatalogAssigneeIdentifier(id="user_3", type="user")
UG_1 = CatalogAssigneeIdentifier(id="ug_1", type="userGroup")
UG_2 = CatalogAssigneeIdentifier(id="ug_2", type="userGroup")
UG_3 = CatalogAssigneeIdentifier(id="ug_3", type="userGroup")

UPSTREAM_PERMISSIONS = [
    CatalogDeclarativeSingleWorkspacePermission(
        name="ANALYZE", assignee=USER_1
    ),
    CatalogDeclarativeSingleWorkspacePermission(name="VIEW", assignee=USER_1),
    CatalogDeclarativeSingleWorkspacePermission(name="MANAGE", assignee=USER_1),
    CatalogDeclarativeSingleWorkspacePermission(
        name="ANALYZE", assignee=USER_2
    ),
    CatalogDeclarativeSingleWorkspacePermission(name="VIEW", assignee=USER_2),
    CatalogDeclarativeSingleWorkspacePermission(
        name="ANALYZE", assignee=USER_3
    ),
    CatalogDeclarativeSingleWorkspacePermission(name="ANALYZE", assignee=UG_1),
    CatalogDeclarativeSingleWorkspacePermission(name="VIEW", assignee=UG_1),
    CatalogDeclarativeSingleWorkspacePermission(name="MANAGE", assignee=UG_1),
    CatalogDeclarativeSingleWorkspacePermission(name="ANALYZE", assignee=UG_2),
    CatalogDeclarativeSingleWorkspacePermission(name="VIEW", assignee=UG_2),
    CatalogDeclarativeSingleWorkspacePermission(name="ANALYZE", assignee=UG_3),
]

WS_PERMISSION_DECLARATION = PermissionDeclaration(
    users={
        "user_1": {"ANALYZE": True, "VIEW": True, "MANAGE": True},
        "user_2": {"ANALYZE": True, "VIEW": True},
        "user_3": {"ANALYZE": True},
    },
    user_groups={
        "ug_1": {"ANALYZE": True, "VIEW": True, "MANAGE": True},
        "ug_2": {"ANALYZE": True, "VIEW": True},
        "ug_3": {"ANALYZE": True},
    },
)

UPSTREAM_WS_PERMISSION = CatalogDeclarativeWorkspacePermissions(
    permissions=UPSTREAM_PERMISSIONS
)

UPSTREAM_WS_PERMISSIONS = {
    "ws_id_1": UPSTREAM_WS_PERMISSION,
    "ws_id_2": UPSTREAM_WS_PERMISSION,
}


def test_declaration_from_populated_sdk_api_obj():
    declaration = PermissionDeclaration.from_sdk_api(UPSTREAM_WS_PERMISSION)
    assert declaration == WS_PERMISSION_DECLARATION


def test_declaration_from_empty_sdk_api_obj():
    api_obj = CatalogDeclarativeWorkspacePermissions(permissions=[])
    declaration = PermissionDeclaration.from_sdk_api(api_obj)
    assert len(declaration.users) == 0
    assert len(declaration.user_groups) == 0


def test_declaration_to_populated_sdk_api_obj():
    api_obj = PermissionDeclaration.to_sdk_api(WS_PERMISSION_DECLARATION)
    assert api_obj == UPSTREAM_WS_PERMISSION


def test_declaration_with_inactive_to_sdk_api_obj():
    users = {
        "user_1": {"ANALYZE": True, "VIEW": False},
        "user_2": {"ANALYZE": True},
    }
    ugs = {
        "ug_1": {"ANALYZE": True, "VIEW": False},
        "ug_2": {"ANALYZE": True},
    }
    declaration = PermissionDeclaration(users, ugs)
    api_obj = declaration.to_sdk_api()
    expected = CatalogDeclarativeWorkspacePermissions(
        permissions=[
            CatalogDeclarativeSingleWorkspacePermission(
                name="ANALYZE", assignee=USER_1
            ),
            CatalogDeclarativeSingleWorkspacePermission(
                name="ANALYZE", assignee=USER_2
            ),
            CatalogDeclarativeSingleWorkspacePermission(
                name="ANALYZE", assignee=UG_1
            ),
            CatalogDeclarativeSingleWorkspacePermission(
                name="ANALYZE", assignee=UG_2
            ),
        ]
    )
    assert api_obj == expected


def test_declaration_with_only_inactive_to_sdk_api_obj():
    users = {
        "user_1": {"ANALYZE": False, "VIEW": False},
        "user_2": {"ANALYZE": False},
    }
    ugs = {
        "ug_1": {"ANALYZE": False, "VIEW": False},
        "ug_2": {"ANALYZE": False},
    }
    declaration = PermissionDeclaration(users, ugs)
    api_obj = declaration.to_sdk_api()
    expected = CatalogDeclarativeWorkspacePermissions(permissions=[])
    assert api_obj == expected


# Declarations are explicitly defined anew here to avoid dict mutations
# in subsequent calls and to avoid dict deepcopy overhead.


def test_add_new_active_user_perm() -> None:
    declaration = PermissionDeclaration(
        {"user_1": {"ANALYZE": True, "VIEW": False}},
        {"ug_1": {"VIEW": True, "ANALYZE": False}},
    )
    permission = PermissionIncrementalLoad(
        permission="MANAGE",
        workspace_id="",
        entity_id="user_1",
        is_active=True,
        entity_type=EntityType.user,
    )

    declaration.add_incremental_permission(permission)
    assert declaration.users == {
        "user_1": {"ANALYZE": True, "VIEW": False, "MANAGE": True}
    }
    assert declaration.user_groups == {"ug_1": {"VIEW": True, "ANALYZE": False}}


def test_add_new_inactive_user_perm() -> None:
    declaration = PermissionDeclaration(
        {"user_1": {"ANALYZE": True, "VIEW": False}},
        {"ug_1": {"VIEW": True, "ANALYZE": False}},
    )
    permission = PermissionIncrementalLoad(
        permission="MANAGE",
        workspace_id="",
        entity_id="user_1",
        entity_type=EntityType.user,
        is_active=False,
    )

    declaration.add_incremental_permission(permission)
    assert declaration.users == {
        "user_1": {"ANALYZE": True, "VIEW": False, "MANAGE": False}
    }
    assert declaration.user_groups == {"ug_1": {"VIEW": True, "ANALYZE": False}}


def test_overwrite_inactive_user_perm() -> None:
    declaration = PermissionDeclaration(
        {"user_1": {"ANALYZE": True, "VIEW": False}},
        {"ug_1": {"VIEW": True, "ANALYZE": False}},
    )
    permission = PermissionIncrementalLoad(
        workspace_id="",
        permission="VIEW",
        entity_id="user_1",
        entity_type=EntityType.user,
        is_active=True,
    )

    declaration.add_incremental_permission(permission)
    assert declaration.users == {"user_1": {"ANALYZE": True, "VIEW": True}}
    assert declaration.user_groups == {"ug_1": {"VIEW": True, "ANALYZE": False}}


def test_overwrite_active_user_perm() -> None:
    declaration = PermissionDeclaration(
        {"user_1": {"ANALYZE": True, "VIEW": False}},
        {"ug_1": {"VIEW": True, "ANALYZE": False}},
    )
    permission = PermissionIncrementalLoad(
        workspace_id="",
        permission="ANALYZE",
        entity_id="user_1",
        entity_type=EntityType.user,
        is_active=False,
    )

    declaration.add_incremental_permission(permission)
    assert declaration.users == {"user_1": {"ANALYZE": True, "VIEW": False}}
    assert declaration.user_groups == {"ug_1": {"VIEW": True, "ANALYZE": False}}


def test_add_new_user_perm() -> None:
    declaration = PermissionDeclaration(
        {"user_1": {"ANALYZE": True, "VIEW": False}},
        {"ug_1": {"VIEW": True, "ANALYZE": False}},
    )

    permission = PermissionIncrementalLoad(
        workspace_id="",
        permission="VIEW",
        entity_id="user_2",
        entity_type=EntityType.user,
        is_active=True,
    )

    declaration.add_incremental_permission(permission)
    assert declaration.users == {
        "user_1": {"ANALYZE": True, "VIEW": False},
        "user_2": {"VIEW": True},
    }
    assert declaration.user_groups == {"ug_1": {"VIEW": True, "ANALYZE": False}}


def test_modify_one_of_user_perms() -> None:
    declaration = PermissionDeclaration(
        {"user_1": {"ANALYZE": True, "VIEW": False}, "user_2": {"VIEW": True}},
        {"ug_1": {"VIEW": True, "ANALYZE": False}},
    )
    permission = PermissionIncrementalLoad(
        workspace_id="",
        permission="MANAGE",
        entity_id="user_1",
        entity_type=EntityType.user,
        is_active=True,
    )
    declaration.add_incremental_permission(permission)
    assert declaration.users == {
        "user_1": {"ANALYZE": True, "VIEW": False, "MANAGE": True},
        "user_2": {"VIEW": True},
    }
    assert declaration.user_groups == {"ug_1": {"VIEW": True, "ANALYZE": False}}


# Add userGroup permission


def test_add_new_active_ug_perm() -> None:
    declaration = PermissionDeclaration(
        {"user_1": {"ANALYZE": True, "VIEW": False}},
        {"ug_1": {"VIEW": True, "ANALYZE": False}},
    )
    permission = PermissionIncrementalLoad(
        workspace_id="",
        permission="MANAGE",
        entity_id="ug_1",
        entity_type=EntityType.user_group,
        is_active=True,
    )
    declaration.add_incremental_permission(permission)
    assert declaration.users == {"user_1": {"ANALYZE": True, "VIEW": False}}
    assert declaration.user_groups == {
        "ug_1": {"VIEW": True, "ANALYZE": False, "MANAGE": True}
    }


def test_add_new_inactive_ug_perm() -> None:
    declaration = PermissionDeclaration(
        {"user_1": {"ANALYZE": True, "VIEW": False}},
        {"ug_1": {"VIEW": True, "ANALYZE": False}},
    )
    permission = PermissionIncrementalLoad(
        workspace_id="",
        permission="MANAGE",
        entity_id="ug_1",
        entity_type=EntityType.user_group,
        is_active=False,
    )
    declaration.add_incremental_permission(permission)
    assert declaration.users == {"user_1": {"ANALYZE": True, "VIEW": False}}
    assert declaration.user_groups == {
        "ug_1": {"VIEW": True, "ANALYZE": False, "MANAGE": False}
    }


def test_overwrite_inactive_ug_perm() -> None:
    declaration = PermissionDeclaration(
        {"user_1": {"ANALYZE": True, "VIEW": False}},
        {"ug_1": {"VIEW": True, "ANALYZE": False}},
    )
    permission = PermissionIncrementalLoad(
        workspace_id="",
        permission="ANALYZE",
        entity_id="ug_1",
        entity_type=EntityType.user_group,
        is_active=True,
    )
    declaration.add_incremental_permission(permission)
    assert declaration.users == {"user_1": {"ANALYZE": True, "VIEW": False}}
    assert declaration.user_groups == {"ug_1": {"VIEW": True, "ANALYZE": True}}


def test_overwrite_active_ug_perm() -> None:
    declaration = PermissionDeclaration(
        {"user_1": {"ANALYZE": True, "VIEW": False}},
        {"ug_1": {"VIEW": True, "ANALYZE": False}},
    )
    permission = PermissionIncrementalLoad(
        workspace_id="",
        permission="VIEW",
        entity_id="ug_1",
        entity_type=EntityType.user_group,
        is_active=True,
    )
    declaration.add_incremental_permission(permission)
    assert declaration.users == {"user_1": {"ANALYZE": True, "VIEW": False}}
    assert declaration.user_groups == {"ug_1": {"VIEW": True, "ANALYZE": False}}


def test_add_new_ug_perm() -> None:
    declaration = PermissionDeclaration(
        {"user_1": {"ANALYZE": True, "VIEW": False}},
        {"ug_1": {"VIEW": True, "ANALYZE": False}},
    )
    permission = PermissionIncrementalLoad(
        workspace_id="",
        permission="VIEW",
        entity_id="ug_2",
        entity_type=EntityType.user_group,
        is_active=True,
    )
    declaration.add_incremental_permission(permission)
    assert declaration.users == {"user_1": {"ANALYZE": True, "VIEW": False}}
    assert declaration.user_groups == {
        "ug_1": {"VIEW": True, "ANALYZE": False},
        "ug_2": {"VIEW": True},
    }


def test_modify_one_of_ug_perms() -> None:
    declaration = PermissionDeclaration(
        {"user_1": {"ANALYZE": True, "VIEW": False}},
        {"ug_1": {"VIEW": True, "ANALYZE": False}, "ug_2": {"VIEW": True}},
    )
    permission = PermissionIncrementalLoad(
        workspace_id="",
        permission="MANAGE",
        entity_id="ug_1",
        entity_type=EntityType.user_group,
        is_active=True,
    )
    declaration.add_incremental_permission(permission)
    assert declaration.users == {"user_1": {"ANALYZE": True, "VIEW": False}}
    assert declaration.user_groups == {
        "ug_1": {"VIEW": True, "ANALYZE": False, "MANAGE": True},
        "ug_2": {"VIEW": True},
    }


def test_upsert() -> None:
    owner = PermissionDeclaration(
        {"user_1": {"ANALYZE": True}, "user_2": {"VIEW": True}},
        {"ug_1": {"ANALYZE": True}, "ug_2": {"VIEW": True}},
    )
    other = PermissionDeclaration(
        {"user_1": {"MANAGE": True, "VIEW": False}},
        {"ug_2": {"MANAGE": True, "VIEW": False}},
    )
    owner.upsert(other)
    assert owner.users == {
        "user_1": {"MANAGE": True, "VIEW": False},
        "user_2": {"VIEW": True},
    }
    assert owner.user_groups == {
        "ug_1": {"ANALYZE": True},
        "ug_2": {"MANAGE": True, "VIEW": False},
    }


def mock_upstream_perms(ws_id: str) -> CatalogDeclarativeWorkspacePermissions:
    if ws_id not in UPSTREAM_WS_PERMISSIONS:
        raise NotFoundException(404)
    return UPSTREAM_WS_PERMISSIONS[ws_id]


def parse_expected_permissions(
    raw_data: dict,
) -> dict[str, list[CatalogDeclarativeSingleWorkspacePermission]]:
    expected_result: dict[
        str, list[CatalogDeclarativeSingleWorkspacePermission]
    ] = {}
    for workspace_id, workspace_permissions in raw_data.items():
        expected_permissions = []
        for permission in workspace_permissions:
            expected_permissions.append(
                CatalogDeclarativeSingleWorkspacePermission(
                    name=permission["name"],
                    assignee=CatalogAssigneeIdentifier(
                        id=permission["assignee_id"],
                        type=permission["assignee_type"],
                    ),
                )
            )
        expected_result[workspace_id] = expected_permissions
    return expected_result


@pytest.mark.parametrize(
    ("source_data_path", "expected_data_path", "load_method"),
    [
        (
            "permissions_input_full_load.json",
            "permissions_expected_full_load.json",
            "full_load",
        ),
        (
            "permissions_input_incremental_load.json",
            "permissions_expected_incremental_load.json",
            "incremental_load",
        ),
    ],
)
def test_permission_provisioner(
    source_data_path: str,
    expected_data_path: str,
    load_method: Literal["incremental_load", "full_load"],
    permission_provisioner: PermissionProvisioner,
    mocker: MockerFixture,
) -> None:
    source_data: list[dict] = []
    incremental_load_data: list[PermissionIncrementalLoad] = []
    full_load_data: list[PermissionFullLoad] = []

    # Load existing upstream permissions
    EXISTING_UPSTREAM_PERMISSIONS_PATH = (
        f"{TEST_DATA_SUBDIR}/existing_upstream_permissions.json"
    )
    with open(EXISTING_UPSTREAM_PERMISSIONS_PATH, "r") as f:
        raw_existing_upstream_permissions = orjson.loads(f.read())

    existing_upstream_permissions = parse_expected_permissions(
        raw_existing_upstream_permissions
    )

    def mock_get_declarative_permissions(
        ws_id: str,
    ) -> CatalogDeclarativeWorkspacePermissions:
        return CatalogDeclarativeWorkspacePermissions(
            permissions=existing_upstream_permissions[ws_id]
        )

    # Patch the get method to return existing upstream permissions
    mocker.patch.object(
        permission_provisioner._api,
        "get_declarative_permissions",
        side_effect=mock_get_declarative_permissions,
    )

    # Load source data
    with open(f"{TEST_DATA_SUBDIR}/{source_data_path}", "r") as f:
        source_data = orjson.loads(f.read())

    # Load and parse expected data
    with open(f"{TEST_DATA_SUBDIR}/{expected_data_path}", "r") as f:
        raw_expected_result = orjson.loads(f.read())

    expected_result = parse_expected_permissions(raw_expected_result)

    # Patch the put method to capture output and compare it with expected result
    def compare_permissions(
        workspace_id: str,
        ws_permissions: CatalogDeclarativeWorkspacePermissions,
    ) -> None:
        actual_permissions = ws_permissions.permissions
        expected_permissions = expected_result[workspace_id]

        assert len(actual_permissions) == len(expected_permissions)

        actual_sorted_permissions = sorted(
            actual_permissions, key=lambda x: x.assignee.id
        )
        expected_sorted_permissions = sorted(
            expected_permissions, key=lambda x: x.assignee.id
        )
        assert actual_sorted_permissions == expected_sorted_permissions

    mocker.patch.object(
        permission_provisioner._api,
        "put_declarative_permissions",
        side_effect=compare_permissions,
    )

    if load_method == "incremental_load":
        incremental_load_data = [
            PermissionIncrementalLoad(**row) for row in source_data
        ]
        permission_provisioner.incremental_load(incremental_load_data)
    else:
        full_load_data = [PermissionFullLoad(**row) for row in source_data]
        permission_provisioner.full_load(full_load_data)



# (C) 2025 GoodData Corporation

"""Utility functions for GoodData Cloud API interactions."""

from typing import Any, Callable

from gooddata_api_client import ApiException  # type: ignore

from gooddata_pipelines.api.exceptions import GoodDataApiException


def raise_with_context(**context_kwargs: str) -> Callable:
    """
    Decorator to catch exceptions raised by SDK methods and raise a GoodDataApiException
    with additional context information.

    Args:
        context_kwargs (dict): Additional context information to include in the
            GoodDataApiException.
    """

    def decorator(fn: Callable) -> Callable:
        def wrapper(*method_args: Any, **method_kwargs: Any) -> Callable:
            try:
                return fn(*method_args, **method_kwargs)
            except Exception as e:
                # Process known exceptions
                if isinstance(e, ApiException):
                    context_kwargs["http_status"] = f"{e.status} {e.reason}"
                    exception_content = e.body
                else:
                    exception_content = str(e)

                # Format the exception message: "{exception_type}: {exception_content}"
                message = f"{type(e).__name__}: {exception_content}"

                raise GoodDataApiException(
                    message, **context_kwargs, **method_kwargs
                )

        return wrapper

    return decorator



# (C) 2025 GoodData Corporation

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    ValidationInfo,
    field_validator,
)


class UserGroupBase(BaseModel):
    model_config = ConfigDict(extra="forbid")

    user_group_id: str
    user_group_name: str
    parent_user_groups: list[str] = Field(default_factory=list)

    @field_validator("user_group_name", mode="before")
    @classmethod
    def validate_user_group_name(
        cls, v: str | None, info: ValidationInfo
    ) -> str:
        """If user_group_name is None or empty, default to user_group_id."""
        if not v:  # handles None and empty string
            return info.data.get("user_group_id", "")
        return v

    @field_validator("parent_user_groups", mode="before")
    @classmethod
    def validate_parent_user_groups(cls, v: list[str] | None) -> list[str]:
        """If parent_user_groups is None or empty, default to empty list."""
        if not v:
            return []
        return v


class UserGroupFullLoad(UserGroupBase):
    """Input validator for full load of user group provisioning."""


class UserGroupIncrementalLoad(UserGroupBase):
    """Input validator for incremental load of user group provisioning."""

    is_active: bool



# (C) 2025 GoodData Corporation

from gooddata_sdk.catalog.workspace.entity_model.workspace import (
    CatalogWorkspace,
)

from gooddata_pipelines.provisioning.entities.workspaces.models import (
    WorkspaceFullLoad,
)
from gooddata_pipelines.provisioning.entities.workspaces.workspace_data_parser import (
    WorkspaceDataParser,
)

parser = WorkspaceDataParser()


def test_get_id_to_name_map_no_overlap() -> None:
    """No overlap between source and Panther groups."""
    source_group: list[WorkspaceFullLoad] = [
        WorkspaceFullLoad(
            parent_id="some_parent",
            workspace_id="1",
            workspace_name="Source Workspace 1",
        ),
        WorkspaceFullLoad(
            parent_id="some_parent",
            workspace_id="2",
            workspace_name="Source Workspace 2",
        ),
    ]
    panther_group: list[CatalogWorkspace] = [
        CatalogWorkspace(workspace_id="3", name="Panther Workspace 1"),
        CatalogWorkspace(workspace_id="4", name="Panther Workspace 2"),
    ]

    expected_result = {
        "3": "Panther Workspace 1",
        "4": "Panther Workspace 2",
        "1": "Source Workspace 1",
        "2": "Source Workspace 2",
    }

    result = parser._get_id_to_name_map(source_group, panther_group)
    assert result == expected_result


def test_get_id_to_name_map_with_overlap() -> None:
    """Overlaping groups -> source group name takse precedence."""
    source_group: list[WorkspaceFullLoad] = [
        WorkspaceFullLoad(
            parent_id="some_parent",
            workspace_id="1",
            workspace_name="Source Workspace 1",
        ),
        WorkspaceFullLoad(
            parent_id="some_parent",
            workspace_id="2",
            workspace_name="Source Workspace 2",
        ),
    ]
    panther_group: list[CatalogWorkspace] = [
        CatalogWorkspace(workspace_id="1", name="Panther Workspace 1"),
        CatalogWorkspace(workspace_id="4", name="Panther Workspace 2"),
    ]

    expected_result = {
        "1": "Source Workspace 1",
        "4": "Panther Workspace 2",
        "2": "Source Workspace 2",
    }

    result = parser._get_id_to_name_map(source_group, panther_group)
    assert result == expected_result


def test_get_id_to_name_map_empty() -> None:
    """Empty source and Panther groups -> will return empty dict."""
    source_group: list[WorkspaceFullLoad] = []
    panther_group: list[CatalogWorkspace] = []

    expected_result: dict[str, str] = {}

    result = parser._get_id_to_name_map(source_group, panther_group)
    assert result == expected_result


def test_get_child_to_parent_map() -> None:
    """Maps child ID to parent ID."""
    source_group: list[WorkspaceFullLoad] = [
        WorkspaceFullLoad(
            workspace_id="child_1",
            parent_id="parent_1",
            workspace_name="Child Workspace 1",
        ),
        WorkspaceFullLoad(
            workspace_id="child_2",
            parent_id="parent_2",
            workspace_name="Child Workspace 2",
        ),
        WorkspaceFullLoad(
            workspace_id="child_3",
            parent_id="parent_1",
            workspace_name="Child Workspace 3",
        ),
    ]

    expected_result = {
        "child_1": "parent_1",
        "child_2": "parent_2",
        "child_3": "parent_1",
    }

    result = parser._get_child_to_parent_map(source_group)
    assert result == expected_result


def test_get_child_to_parent_map_empty() -> None:
    """Empty source group -> will return empty dict."""
    source_group: list[WorkspaceFullLoad] = []

    expected_result: dict[str, str] = {}

    result = parser._get_child_to_parent_map(source_group)
    assert result == expected_result


def test_get_child_to_parent_map_with_duplicates() -> None:
    """Child ID will be unique in the resulting dict"""
    source_group: list[WorkspaceFullLoad] = [
        WorkspaceFullLoad(
            workspace_id="child_1",
            parent_id="parent_1",
            workspace_name="Child Workspace 1",
        ),
        WorkspaceFullLoad(
            workspace_id="child_2",
            parent_id="parent_2",
            workspace_name="Child Workspace 2",
        ),
        WorkspaceFullLoad(
            workspace_id="child_3",
            parent_id="parent_1",
            workspace_name="Child Workspace 3",
        ),
        WorkspaceFullLoad(
            workspace_id="child_1",
            parent_id="parent_1",
            workspace_name="Child Workspace 1",
        ),
        WorkspaceFullLoad(
            workspace_id="child_2",
            parent_id="parent_2",
            workspace_name="Child Workspace 2",
        ),
        WorkspaceFullLoad(
            workspace_id="child_3",
            parent_id="parent_1",
            workspace_name="Child Workspace 3",
        ),
    ]

    expected_result: dict[str, str] = {
        "child_1": "parent_1",
        "child_2": "parent_2",
        "child_3": "parent_1",
    }

    result = parser._get_child_to_parent_map(source_group)
    assert result == expected_result


def test_get_child_to_wdfs_map() -> None:
    """Mapping child ID to WDF ID and WDF values."""
    source_group: list[WorkspaceFullLoad] = [
        WorkspaceFullLoad(
            parent_id="parent_1",
            workspace_id="child_1",
            workspace_name="Child Workspace 1",
            workspace_data_filter_id="wdf_1",
            workspace_data_filter_values=["value_1", "value_2"],
        ),
        WorkspaceFullLoad(
            parent_id="parent_2",
            workspace_id="child_2",
            workspace_name="Child Workspace 2",
            workspace_data_filter_id="wdf_2",
            workspace_data_filter_values=["value_3", "value_4"],
        ),
        WorkspaceFullLoad(
            parent_id="parent_1",
            workspace_id="child_3",
            workspace_name="Child Workspace 3",
            workspace_data_filter_id=None,
            workspace_data_filter_values=None,
        ),
    ]

    expected_result = {
        "child_1": {"wdf_1": ["value_1", "value_2"]},
        "child_2": {"wdf_2": ["value_3", "value_4"]},
    }

    result = parser._get_child_to_wdfs_map(source_group)
    assert result == expected_result


def test_get_child_to_wdfs_map_integers() -> None:
    """Is capable of handling int values (in case source column is int)."""
    source_group: list[WorkspaceFullLoad] = [
        WorkspaceFullLoad(
            workspace_id="child_1",
            parent_id="parent_1",
            workspace_name="Child Workspace 1",
            workspace_data_filter_id="wdf_1",
            workspace_data_filter_values=[1],  # type: ignore
        ),
        WorkspaceFullLoad(
            workspace_id="child_2",
            parent_id="parent_2",
            workspace_name="Child Workspace 2",
            workspace_data_filter_id="wdf_2",
            workspace_data_filter_values=["value_3", "value_4"],
        ),
        WorkspaceFullLoad(
            workspace_id="child_3",
            parent_id="parent_1",
            workspace_name="Child Workspace 3",
            workspace_data_filter_id=None,
            workspace_data_filter_values=None,
        ),
    ]

    expected_result = {
        "child_1": {"wdf_1": ["1"]},
        "child_2": {"wdf_2": ["value_3", "value_4"]},
    }

    result = parser._get_child_to_wdfs_map(source_group)
    assert result == expected_result


def test_get_child_to_wdfs_map_multiple_wdfs() -> None:
    """Handles multiple WDFs on a child workspace."""
    source_group: list[WorkspaceFullLoad] = [
        WorkspaceFullLoad(
            workspace_id="child_1",
            parent_id="parent_1",
            workspace_name="Child Workspace 1",
            workspace_data_filter_id="wdf_1",
            workspace_data_filter_values=[1],  # type: ignore
        ),
        WorkspaceFullLoad(
            workspace_id="child_1",
            parent_id="parent_1",
            workspace_name="Child Workspace 1",
            workspace_data_filter_id="wdf_2",
            workspace_data_filter_values=["value_3", "value_4"],
        ),
        WorkspaceFullLoad(
            workspace_id="child_2",
            parent_id="parent_2",
            workspace_name="Child Workspace 2",
            workspace_data_filter_id="wdf_2",
            workspace_data_filter_values=["value_3", "value_4"],
        ),
        WorkspaceFullLoad(
            workspace_id="child_3",
            parent_id="parent_1",
            workspace_name="Child Workspace 3",
            workspace_data_filter_id=None,
            workspace_data_filter_values=None,
        ),
    ]

    expected_result = {
        "child_1": {"wdf_1": ["1"], "wdf_2": ["value_3", "value_4"]},
        "child_2": {"wdf_2": ["value_3", "value_4"]},
    }

    result = parser._get_child_to_wdfs_map(source_group)
    assert result == expected_result


def test_get_child_to_wdfs_map_empty() -> None:
    source_group: list[WorkspaceFullLoad] = []

    expected_result: dict[str, dict[str, list[str]]] = {}

    result = parser._get_child_to_wdfs_map(source_group)
    assert result == expected_result



# (C) 2025 GoodData Corporation

"""Interaction with GoodData Cloud via the Gooddata Python SDK."""

from pathlib import Path
from typing import Callable

from gooddata_sdk.catalog.permission.declarative_model.permission import (
    CatalogDeclarativeWorkspacePermissions,
)
from gooddata_sdk.catalog.user.entity_model.user import CatalogUser
from gooddata_sdk.catalog.user.entity_model.user_group import CatalogUserGroup
from gooddata_sdk.catalog.workspace.declarative_model.workspace.workspace import (
    CatalogDeclarativeWorkspaceDataFilters,
)
from gooddata_sdk.catalog.workspace.entity_model.user_data_filter import (
    CatalogUserDataFilter,
)
from gooddata_sdk.catalog.workspace.entity_model.workspace import (
    CatalogWorkspace,
)
from gooddata_sdk.sdk import GoodDataSdk

from gooddata_pipelines.api.utils import raise_with_context


def apply_to_all_methods(decorator: Callable) -> Callable:
    def decorate(cls: type) -> type:
        for attr in cls.__dict__:
            if callable(getattr(cls, attr)) and not attr.startswith("__"):
                setattr(cls, attr, decorator(getattr(cls, attr)))
        return cls

    return decorate


@apply_to_all_methods(raise_with_context())
class SdkMethods:
    """
    Class to intaract with GoodData Cloud via the Gooddata Python SDK.
    """

    _sdk: GoodDataSdk

    def get_organization_id(self) -> str:
        return self._sdk.catalog_organization.organization_id

    def check_workspace_exists(self, workspace_id: str) -> bool:
        try:
            self._sdk.catalog_workspace.get_workspace(workspace_id)
            return True
        except Exception:
            return False

    def get_workspace(self, workspace_id: str, **_: str) -> CatalogWorkspace:
        """
        Calls GoodData Python SDK to retrieve a workspace by its ID.

        Args:
            workspace_id (str): The ID of the workspace to retrieve.
        Returns:
            CatalogWorkspace: The workspace object retrieved from the SDK.
        Raises:
            GoodDataApiException: If the workspace cannot be retrieved, an exception
            is raised with additional context information.
        """
        return self._sdk.catalog_workspace.get_workspace(workspace_id)

    def delete_panther_workspace(self, workspace_id: str) -> None:
        """
        Calls GoodData Python SDK to delete a workspace by its ID.

        Args:
            workspace_id (str): The ID of the workspace to delete.
        Raises:
            GoodDataApiException: If the workspace cannot be deleted, an exception
            is raised with additional context information.
        """
        self._sdk.catalog_workspace.delete_workspace(workspace_id)

    def create_or_update_panther_workspace(
        self,
        workspace_id: str,
        workspace_name: str,
        parent_id: str | None,
        **_: str,
    ) -> None:
        """
        Calls GoodData Python SDK to create or update a workspace with the given ID,
        name, and parent ID.

        Args:
            workspace_id (str): The ID of the workspace to create or update.
            workspace_name (str): The name of the workspace.
            parent_id (str | None): The ID of the parent workspace, if any.
        Returns:
            None
        Raises:
            GoodDataApiException: If the workspace cannot be created or updated,
            an exception is raised with additional context information.
        """
        return self._sdk.catalog_workspace.create_or_update(
            CatalogWorkspace(
                workspace_id=workspace_id,
                name=workspace_name,
                parent_id=parent_id,
            )
        )

    def get_panther_children_workspaces(
        self, parent_workspace_ids: set[str]
    ) -> list[CatalogWorkspace]:
        """
        Calls GoodData Python SDK to retrieve all workspaces in domain and filters the
        result by the set of parent workspace IDs.

        Args:
            parent_workspace_ids (set[str]): A set of parent workspace IDs to filter
                child workspaces.
        Returns:
            list[CatalogWorkspace]: List of child workspaces in the parent workspace.
        """
        all_workspaces: list[CatalogWorkspace] = self.list_workspaces()

        children: list[CatalogWorkspace] = [
            workspace
            for workspace in all_workspaces
            if workspace.parent_id in parent_workspace_ids
        ]

        return children

    def list_workspaces(self) -> list[CatalogWorkspace]:
        """Retrieves all workspaces in the GoodData Cloud domain.

        Returns:
            list[CatalogWorkspace]: A list of all workspaces in the domain.
        Raises:
            GoodDataApiException: If the workspaces cannot be retrieved, an exception
            is raised with additional context information.
        """
        return self._sdk.catalog_workspace.list_workspaces()

    def get_declarative_permissions(
        self, workspace_id: str
    ) -> CatalogDeclarativeWorkspacePermissions:
        """
        Retrieves the declarative permissions for a given workspace.

        Args:
            workspace_id (str): The ID of the workspace for which to retrieve
                permissions.
        Returns:
            CatalogDeclarativeWorkspacePermissions: The declarative permissions
                for the workspace.
        Raises:
            GoodDataApiException: If the permissions cannot be retrieved, an exception
            is raised with additional context information.
        """
        return self._sdk.catalog_permission.get_declarative_permissions(
            workspace_id
        )

    def put_declarative_permissions(
        self,
        workspace_id: str,
        ws_permissions: CatalogDeclarativeWorkspacePermissions,
    ) -> None:
        """
        Updates the declarative permissions for a given workspace.

        Args:
            workspace_id (str): The ID of the workspace for which to update
                permissions.
            ws_permissions (CatalogDeclarativeWorkspacePermissions): The new
                declarative permissions to set for the workspace.
        Returns:
            None
        Raises:
            GoodDataApiException: If the permissions cannot be updated, an exception
            is raised with additional context information.
        """
        return self._sdk.catalog_permission.put_declarative_permissions(
            workspace_id, ws_permissions
        )

    def get_user(self, user_id: str, **_: str) -> CatalogUser:
        """
        Calls GoodData Python SDK to retrieve a user by its ID.

        Args:
            user_id (str): The ID of the user to retrieve.
        Returns:
            CatalogUser: The user object retrieved from the SDK.
        Raises:
            GoodDataApiException: If the user cannot be retrieved, an exception
            is raised with additional context information.
        """
        return self._sdk.catalog_user.get_user(user_id)

    def create_or_update_user(self, user: CatalogUser, **_: str) -> None:
        """
        Calls GoodData Python SDK to create or update a user.

        Args:
            user (CatalogUser): The user object to create or update.
        Returns:
            None
        Raises:
            GoodDataApiException: If the user cannot be created or updated,
            an exception is raised with additional context information.
        """
        return self._sdk.catalog_user.create_or_update_user(user)

    def delete_user(self, user_id: str, **_: str) -> None:
        """
        Calls GoodData Python SDK to delete a user by its ID.

        Args:
            user_id (str): The ID of the user to delete.
        Returns:
            None
        Raises:
            GoodDataApiException: If the user cannot be deleted, an exception
            is raised with additional context information.
        """
        return self._sdk.catalog_user.delete_user(user_id)

    def get_user_group(self, user_group_id: str, **_: str) -> CatalogUserGroup:
        """
        Calls GoodData Python SDK to retrieve a user group by its ID.

        Args:
            user_group_id (str): The ID of the user group to retrieve.
        Returns:
            CatalogUserGroup: The user group object retrieved from the SDK.
        Raises:
            GoodDataApiException: If the user group cannot be retrieved, an exception
            is raised with additional context information.
        """
        return self._sdk.catalog_user.get_user_group(user_group_id)

    def list_user_groups(self) -> list[CatalogUserGroup]:
        """
        Calls GoodData Python SDK to retrieve all user groups.

        Returns:
            list[CatalogUserGroup]: A list of all user groups in the domain.
        Raises:
            GoodDataApiException: If the user groups cannot be retrieved, an
            exception is raised with additional context information.
        """
        return self._sdk.catalog_user.list_user_groups()

    def list_users(self) -> list[CatalogUser]:
        """Calls GoodData Python SDK to retrieve all users.

        Returns:
            list[CatalogUser]: A list of all users in the domain.
        """
        return self._sdk.catalog_user.list_users()

    def create_or_update_user_group(
        self, catalog_user_group: CatalogUserGroup, **_: str
    ) -> None:
        """Calls GoodData Python SDK to create or update a user group.

        Args:
            catalog_user_group (CatalogUserGroup): The user group object to create or update.
        Returns:
            None
        Raises:
            GoodDataApiException: If the user group cannot be created or updated,
            an exception is raised with additional context information.
        """
        return self._sdk.catalog_user.create_or_update_user_group(
            catalog_user_group
        )

    def delete_user_group(self, user_group_id: str) -> None:
        """Calls GoodData Python SDK to delete a user group by its ID.

        Args:
            user_group_id (str): The ID of the user group to delete.
        Returns:
            None
        Raises:
            GoodDataApiException: If the user group cannot be deleted, an exception
            is raised with additional context information.
        """
        return self._sdk.catalog_user.delete_user_group(user_group_id)

    def get_declarative_workspace_data_filters(
        self,
    ) -> CatalogDeclarativeWorkspaceDataFilters:
        """Retrieves the declarative workspace data filters.

        Returns:
            CatalogDeclarativeWorkspaceDataFilters: The declarative workspace data filters.
        Raises:
            GoodDataApiException: If the declarative workspace data filters cannot be retrieved,
            an exception is raised with additional context information.
        """
        return (
            self._sdk.catalog_workspace.get_declarative_workspace_data_filters()
        )

    def list_user_data_filters(
        self, workspace_id: str
    ) -> list[CatalogUserDataFilter]:
        """Lists all user data filters for a given workspace.

        Args:
            workspace_id (str): The ID of the workspace for which to list user data filters.
        Returns:
            list[CatalogUserDataFilter]: A list of user data filters for the specified workspace.
        Raises:
            GoodDataApiException: If the user data filters cannot be listed, an exception
            is raised with additional context information.
        """
        return self._sdk.catalog_workspace.list_user_data_filters(workspace_id)

    def delete_user_data_filter(
        self, workspace_id: str, user_data_filter_id: str
    ) -> None:
        """Deletes a user data filter by its ID in the specified workspace.

        Args:
            workspace_id (str): The ID of the workspace containing the user data filter.
            user_data_filter_id (str): The ID of the user data filter to delete.
        Returns:
            None
        Raises:
            GoodDataApiException: If the user data filter cannot be deleted, an exception
            is raised with additional context information.
        """
        self._sdk.catalog_workspace.delete_user_data_filter(
            workspace_id, user_data_filter_id
        )

    def create_or_update_user_data_filter(
        self, workspace_id: str, user_data_filter: CatalogUserDataFilter
    ) -> None:
        """Creates or updates a user data filter in the specified workspace.

        Args:
            workspace_id (str): The ID of the workspace where the user data filter
                should be created or updated.
            user_data_filter (CatalogUserDataFilter): The user data filter object to create or update.
        Returns:
            None
        Raises:
            GoodDataApiException: If the user data filter cannot be created or updated,
            an exception is raised with additional context information.
        """
        self._sdk.catalog_workspace.create_or_update_user_data_filter(
            workspace_id, user_data_filter
        )

    def store_declarative_workspace(
        self, workspace_id: str, export_path: Path
    ) -> None:
        """Stores the declarative workspace in the specified export path."""
        self._sdk.catalog_workspace.store_declarative_workspace(
            workspace_id, export_path
        )

    def store_declarative_filter_views(
        self, workspace_id: str, export_path: Path
    ) -> None:
        """Stores the declarative filter views in the specified export path."""
        self._sdk.catalog_workspace.store_declarative_filter_views(
            workspace_id, export_path
        )



# (C) 2025 GoodData Corporation

import time
import threading
import functools
from typing import Callable, Any, Literal


class RateLimiter:
    """
    Rate limiter usable as a decorator and as a context manager.
      - Shared instance decorator:   limiter = RateLimiter(); @limiter
      - Per-function decorator:      @RateLimiter(calls_per_second=2)
      - Context manager:             with RateLimiter(2): ...
    """

    def __init__(self, calls_per_second: float = 1.0) -> None:
        if calls_per_second <= 0:
            raise ValueError("calls_per_second must be greater than 0")

        self.calls_per_second = calls_per_second
        self.min_interval = 1.0 / calls_per_second

        self._lock = threading.Lock()
        self._last_call_time = 0.0

    def wait_if_needed(self) -> float:
        """Sleep if needed to maintain the rate limit, return actual sleep time."""
        with self._lock:
            now = time.monotonic()
            since_last = now - self._last_call_time

            if since_last < self.min_interval:
                sleep_time = self.min_interval - since_last
                time.sleep(sleep_time)
                self._last_call_time = time.monotonic()
                return sleep_time
            else:
                self._last_call_time = now
                return 0.0

    # Decorator support
    def __call__(self, func: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            self.wait_if_needed()
            return func(*args, **kwargs)

        return wrapper

    # Context manager support
    def __enter__(self) -> "RateLimiter":
        self.wait_if_needed()
        return self

    def __exit__(
        self, exc_type: Any, exc_val: Any, exc_tb: Any
    ) -> Literal[False]:
        return False

    def reset(self) -> None:
        """Reset the limiter (useful in tests)."""
        with self._lock:
            self._last_call_time = 0.0



# (C) 2025 GoodData Corporation

import os
import tempfile
import zipfile
from pathlib import Path
from typing import Any

import attrs
from gooddata_sdk.catalog.workspace.declarative_model.workspace.analytics_model.analytics_model import (
    CatalogDeclarativeAnalytics,
)
from gooddata_sdk.catalog.workspace.declarative_model.workspace.automation import (
    CatalogDeclarativeAutomation,
)
from gooddata_sdk.catalog.workspace.declarative_model.workspace.logical_model.ldm import (
    CatalogDeclarativeModel,
)
from gooddata_sdk.catalog.workspace.declarative_model.workspace.workspace import (
    CatalogDeclarativeFilterView,
)
from pydantic import BaseModel, ConfigDict

from gooddata_pipelines.backup_and_restore.base_manager import BaseManager
from gooddata_pipelines.backup_and_restore.constants import DirNames
from gooddata_pipelines.utils.decorators import log_and_reraise_exception


@attrs.define
class WorkspaceModel:
    logical_data_model: CatalogDeclarativeModel
    analytics_model: CatalogDeclarativeAnalytics


class WorkspaceToRestore(BaseModel):
    """Workspace to restore.

    Args:
        id: The ID of the workspace to restore.
        path: The path to the folder containing the `gooddata_layouts.zip` file
            to restore. Should be a continuation of the `backup_path` specified
            in the storage configuration. Typically, it would look something like
            `organization_id/workspace_id/backup_timestamp`
    """

    model_config = ConfigDict(extra="forbid")

    id: str
    path: str


class RestoreManager(BaseManager):
    """Restores previsouly created backups of workspace metadata."""

    @log_and_reraise_exception("Failed to extract backup from zip archive.")
    def _extract_zip_archive(
        self, file_to_extract: Path, destination: Path
    ) -> None:
        """Extracts the backup from zip archive."""
        with zipfile.ZipFile(file_to_extract, "r") as zip_ref:
            zip_ref.extractall(destination)

    def _check_workspace_is_valid(self, workspace_root_dir_path: Path) -> None:
        """Checks if the workspace layout is valid."""
        if (
            not workspace_root_dir_path.exists()
            or not workspace_root_dir_path.is_dir()
        ):
            self.logger.error(
                "Invalid source path found upon backup fetch. "
                f"Got {workspace_root_dir_path}. "
                "Check if target zip contains gooddata_layouts directory."
            )
            raise RuntimeError("Invalid source path upon load.")

        children = list(workspace_root_dir_path.iterdir())
        am_path = workspace_root_dir_path / DirNames.AM
        ldm_path = workspace_root_dir_path / DirNames.LDM
        udf_path = workspace_root_dir_path / DirNames.UDF

        if (
            am_path not in children
            or ldm_path not in children
            or udf_path not in children
        ):
            self.logger.error(
                f"{DirNames.AM} or {DirNames.LDM} directory missing in the "
                + "workspace hierarchy. "
            )
            raise RuntimeError(
                f"{DirNames.AM} or {DirNames.LDM} directory missing."
            )

    @log_and_reraise_exception("Failed to load workspace declaration.")
    def _load_workspace_layout(
        self, workspace_root_dir_path: Path
    ) -> WorkspaceModel:
        """Loads the workspace layout from the backup."""
        sdk_catalog = self._api._sdk.catalog_workspace_content

        ldm = sdk_catalog.load_ldm_from_disk(workspace_root_dir_path)
        am = sdk_catalog.load_analytics_model_from_disk(workspace_root_dir_path)

        return WorkspaceModel(logical_data_model=ldm, analytics_model=am)

    @log_and_reraise_exception("Failed to load user data filters from folder.")
    def _load_user_data_filters(self, workspace_root_dir_path: Path) -> dict:
        user_data_filters: dict = {"userDataFilters": []}
        user_data_filters_folder = os.path.join(
            workspace_root_dir_path, DirNames.UDF
        )
        for filename in os.listdir(user_data_filters_folder):
            file_path = os.path.join(user_data_filters_folder, filename)
            user_data_filter = self.yaml_utils.safe_load(Path(file_path))
            user_data_filters["userDataFilters"].append(user_data_filter)

        return user_data_filters

    @log_and_reraise_exception("Failed to put workspace layout into GoodData.")
    def _put_workspace_layout(
        self, workspace_id: str, workspace_model: WorkspaceModel
    ) -> None:
        """Puts the workspace layout into GoodData."""
        self._api._sdk.catalog_workspace_content.put_declarative_ldm(
            workspace_id, workspace_model.logical_data_model
        )
        self._api._sdk.catalog_workspace_content.put_declarative_analytics_model(
            workspace_id, workspace_model.analytics_model
        )

    @log_and_reraise_exception("Failed to put user data filters.")
    def _put_user_data_filters(
        self, workspace_id: str, user_data_filters: dict
    ) -> None:
        """Puts the user data filters into GoodData workspace."""
        response = self._api.put_user_data_filters(
            workspace_id, user_data_filters
        )
        self._api.raise_if_response_not_ok(response)

    def _load_and_put_filter_views(
        self, workspace_id: str, workspace_root_dir_path: Path
    ) -> None:
        """Loads and puts filter views into GoodData workspace."""
        filter_views: list[CatalogDeclarativeFilterView] = []
        if not (workspace_root_dir_path / "filter_views").exists():
            # Skip if the filter_views directory does not exist
            return

        for file in Path(workspace_root_dir_path / "filter_views").iterdir():
            filter_view_content: dict[str, Any] = dict(
                self.yaml_utils.safe_load(file)
            )
            filter_view: CatalogDeclarativeFilterView = (
                CatalogDeclarativeFilterView.from_dict(filter_view_content)
            )
            filter_views.append(filter_view)

        if filter_views:
            self._api._sdk.catalog_workspace.put_declarative_filter_views(
                workspace_id, filter_views
            )

    def _load_and_post_automations(
        self, workspace_id: str, workspace_root_dir_path: Path
    ) -> None:
        """Loads automations from specified json file and creates them in the workspace."""
        # Load automations from JSON
        path_to_json: Path = Path(
            workspace_root_dir_path, "automations", "automations.json"
        )

        # Both the folder and the file must exist, otherwise skip
        if not (workspace_root_dir_path.exists() and path_to_json.exists()):
            return

        # Delete all automations from the workspace and restore the automations from the backup.
        self._delete_all_automations(workspace_id)

        data: dict = self.json_utils.load(path_to_json)
        automations: list[dict] = data["data"]

        for automation in automations:
            self._post_automation(workspace_id, automation)

    def _delete_all_automations(self, workspace_id: str) -> None:
        """Deletes all automations in the workspace."""
        automations: list[CatalogDeclarativeAutomation] = (
            self._api._sdk.catalog_workspace.get_declarative_automations(
                workspace_id
            )
        )
        for automation in automations:
            self._api.delete_automation(workspace_id, automation.id)

    def _post_automation(self, workspace_id: str, automation: dict) -> None:
        """Posts a scheduled export to the workspace."""
        attributes: dict = automation["attributes"]
        relationships: dict = automation["relationships"]
        id: str = automation["id"]

        if attributes.get("schedule"):
            if attributes["schedule"].get("cronDescription"):
                # The cron description attribute is causing a 500 ("No mapping found...")
                # error. Known and reported issue.
                del attributes["schedule"]["cronDescription"]

        data = {
            "data": {
                "attributes": attributes,
                "id": id,
                "type": "automation",
                "relationships": relationships,
            }
        }

        response = self._api.post_automation(workspace_id, data)

        if not response.ok:
            self.logger.error(
                f"Failed to post automation ({response.status_code}): {response.text}"
            )

    def _restore_backup(
        self, workspace_to_restore: WorkspaceToRestore, tempdir_path: Path
    ) -> None:
        """Restores the backup of a workspace."""

        zip_target = tempdir_path / f"{DirNames.LAYOUTS}.zip"
        src_path = tempdir_path / DirNames.LAYOUTS

        try:
            self.storage.get_ws_declaration(
                workspace_to_restore.path, str(zip_target)
            )
            self._extract_zip_archive(zip_target, tempdir_path)
            self._check_workspace_is_valid(src_path)
            workspace_model: WorkspaceModel = self._load_workspace_layout(
                src_path
            )
            user_data_filters = self._load_user_data_filters(src_path)
            self._put_workspace_layout(workspace_to_restore.id, workspace_model)
            self._put_user_data_filters(
                workspace_to_restore.id, user_data_filters
            )
            self._load_and_put_filter_views(workspace_to_restore.id, src_path)
            self._load_and_post_automations(workspace_to_restore.id, src_path)
            self.logger.info(
                f"Finished backup restore of {workspace_to_restore.id} from {workspace_to_restore.path}."
            )
        except Exception as e:
            self.logger.error(
                f"Failed to restore backup of {workspace_to_restore.id} from {workspace_to_restore.path}. "
                f"Error caused by {e.__class__.__name__}: {e}."
            )

    def restore(self, workspaces_to_restore: list[WorkspaceToRestore]) -> None:
        """Restores the backups of workspaces.

        Args:
            workspaces_to_restore: List of workspaces to restore.
        """
        for workspace_to_restore in workspaces_to_restore:
            with tempfile.TemporaryDirectory() as tempdir:
                tempdir_path = Path(tempdir)
                self._restore_backup(workspace_to_restore, tempdir_path)



# (C) 2025 GoodData Corporation

"""Module for provisioning user groups in GoodData workspaces."""

from typing import Sequence, TypeAlias

from gooddata_sdk.catalog.user.entity_model.user_group import CatalogUserGroup

from gooddata_pipelines.provisioning.entities.users.models.user_groups import (
    UserGroupFullLoad,
    UserGroupIncrementalLoad,
)
from gooddata_pipelines.provisioning.provisioning import Provisioning

UserGroupModel: TypeAlias = UserGroupFullLoad | UserGroupIncrementalLoad


class UserGroupProvisioner(
    Provisioning[UserGroupFullLoad, UserGroupIncrementalLoad]
):
    """Provisioning class for user groups in GoodData workspaces.

    This class handles the creation, update, and deletion of user groups
    based on the provided source data. Use the `full_load` or `incremental_load`
    methods to run the provisioning.
    """

    source_group_incremental: list[UserGroupIncrementalLoad]
    source_group_full: list[UserGroupFullLoad]
    upstream_user_groups: list[CatalogUserGroup]

    FULL_LOAD_TYPE: type[UserGroupFullLoad] = UserGroupFullLoad
    INCREMENTAL_LOAD_TYPE: type[UserGroupIncrementalLoad] = (
        UserGroupIncrementalLoad
    )

    @staticmethod
    def _is_changed(
        group: UserGroupModel, existing_group: CatalogUserGroup
    ) -> bool:
        """Checks if user group has some changes and needs to be updated."""
        group.parent_user_groups.sort()
        parents_changed = group.parent_user_groups != existing_group.get_parents
        name_changed = group.user_group_name != existing_group.name
        return parents_changed or name_changed

    def _create_or_update_user_group(
        self,
        group_id: str,
        group_name: str,
        parent_user_groups: list[str],
    ) -> None:
        """Creates or updates user group in the project."""
        catalog_user_group = CatalogUserGroup.init(
            user_group_id=group_id,
            user_group_name=group_name,
            user_group_parent_ids=parent_user_groups,
        )
        try:
            self._api.create_or_update_user_group(
                catalog_user_group=catalog_user_group
            )
            self.logger.info(
                f"Created/Updated user group: {group_id} - {group_name}"
            )
        except Exception as e:
            self.logger.error(
                f"Failed to create/update user group. Error: {e} "
                + f"Context: {catalog_user_group.__dict__}"
            )

    def _create_missing_user_groups(
        self,
        groups_to_create: Sequence[UserGroupModel],
    ) -> None:
        """Provisions user groups that don't exist."""
        # Sort user groups to create those without parents first
        sorted_groups = sorted(
            groups_to_create, key=lambda x: 1 if x.parent_user_groups else 0
        )

        for group in sorted_groups:
            self._create_or_update_user_group(
                group.user_group_id,
                group.user_group_name,
                group.parent_user_groups,
            )

    def _update_existing_user_groups(
        self,
        groups_to_update: Sequence[UserGroupModel],
        upstream_user_groups: list[CatalogUserGroup],
    ) -> None:
        """Update existing user groups and update ws_permissions."""
        existing_groups = {group.id: group for group in upstream_user_groups}

        for group in groups_to_update:
            existing_group = existing_groups[group.user_group_id]
            if self._is_changed(group, existing_group):
                self._create_or_update_user_group(
                    group.user_group_id,
                    group.user_group_name,
                    group.parent_user_groups,
                )

    def _delete_user_group(self, group_ids_to_delete: set[str]) -> None:
        """Deletes user group from the project."""
        for group_id in group_ids_to_delete:
            try:
                self._api.delete_user_group(group_id)
                self.logger.info(f"Deleted user group: {group_id}")
            except Exception as e:
                self.logger.error(
                    f"Failed to delete user group. Error: {e} "
                    + f"Context: {{'user_group_id': {group_id}}}"
                )

    def _provision_incremental_load(self) -> None:
        """Runs incremental provisioning of user groups."""
        # Get existing user groups from GoodData Cloud
        self.upstream_user_groups = self._api.list_user_groups()

        # Create a set of upstream user group IDs
        upstream_group_ids: set[str] = {
            group.id for group in self.upstream_user_groups
        }

        # Create a set of active source user group IDs
        active_source_groups: set[str] = {
            group.user_group_id
            for group in self.source_group_incremental
            if group.is_active is True
        }

        # Create a set of inactive source user group IDs
        inactive_source_groups: set[str] = {
            group.user_group_id
            for group in self.source_group_incremental
            if group.is_active is False
        }

        # Create a set of user group IDs to create as the difference between active
        # source groups and upstream groups - i.e, we are creating groups marked
        # as active in the source data but which are missing upstream in GoodData Cloud.
        group_ids_to_create: set[str] = active_source_groups.difference(
            upstream_group_ids
        )

        # Create a set of user group IDs to update as the intersection between active
        # source groups and upstream groups - i.e, we are updating groups marked
        # as active in the source data and which are present upstream in GoodData Cloud.
        # The `_update_existing_user_groups` method will check if the upstream group
        # definition differs from the source and if so, it will update the group.
        group_ids_to_update: set[str] = active_source_groups.intersection(
            upstream_group_ids
        )

        # Create a set of user group IDs to delete as the intersection between
        # inactive source groups and upstream groups - i.e, we are deleting groups
        # marked as inactive in the source data and which are present upstream in
        # GoodData Cloud.
        group_ids_to_delete: set[str] = inactive_source_groups.intersection(
            upstream_group_ids
        )

        # create lists of groups to create, update and delete based on the sets
        groups_to_create: list[UserGroupIncrementalLoad] = []
        groups_to_update: list[UserGroupIncrementalLoad] = []

        for group in self.source_group_incremental:
            if group.user_group_id in group_ids_to_create:
                groups_to_create.append(group)
            elif group.user_group_id in group_ids_to_update:
                groups_to_update.append(group)

        self._create_missing_user_groups(groups_to_create)
        self._update_existing_user_groups(
            groups_to_update, self.upstream_user_groups
        )
        self._delete_user_group(group_ids_to_delete)

    def _provision_full_load(self) -> None:
        """Runs full load provisioning of user groups."""
        # Get upsream user groups
        self.upstream_user_groups = self._api.list_user_groups()

        # Create a set of upstream user group IDs
        upstream_group_ids: set[str] = {
            group.id for group in self.upstream_user_groups
        }

        # Create a set of source user group IDs
        source_group_ids: set[str] = {
            group.user_group_id for group in self.source_group_full
        }

        # Figure out which ids are to be created, deleted or exist in both systems
        id_groups = self._create_groups(source_group_ids, upstream_group_ids)

        groups_to_create: list[UserGroupFullLoad] = []
        groups_to_update: list[UserGroupFullLoad] = []

        for group in self.source_group_full:
            if group.user_group_id in id_groups.ids_to_create:
                groups_to_create.append(group)
            elif group.user_group_id in id_groups.ids_in_both_systems:
                groups_to_update.append(group)

        # Create user groups
        self._create_missing_user_groups(groups_to_create)

        # Update user groups
        self._update_existing_user_groups(
            groups_to_update, self.upstream_user_groups
        )

        # Delete user groups
        self._delete_user_group(id_groups.ids_to_delete)



# (C) 2025 GoodData Corporation

"""Module for provisioning users in GoodData workspaces."""

from typing import TypeAlias

from gooddata_api_client.exceptions import NotFoundException  # type: ignore
from gooddata_sdk.catalog.user.entity_model.user import CatalogUser
from gooddata_sdk.catalog.user.entity_model.user_group import CatalogUserGroup

from gooddata_pipelines.provisioning.entities.users.models.users import (
    UserFullLoad,
    UserIncrementalLoad,
    UserProfile,
)
from gooddata_pipelines.provisioning.provisioning import Provisioning
from gooddata_pipelines.provisioning.utils.context_objects import UserContext

# Type alias for user model instances
UserModel: TypeAlias = UserFullLoad | UserIncrementalLoad
UserId: TypeAlias = str


class UserProvisioner(Provisioning[UserFullLoad, UserIncrementalLoad]):
    """Provisioning class for users in GoodData workspaces.

    This class handles the creation, update, and deletion of users
    based on the provided source data.
    """

    source_group_incremental: list[UserIncrementalLoad]
    source_group_full: list[UserFullLoad]

    current_user_id: str

    FULL_LOAD_TYPE: type[UserFullLoad] = UserFullLoad
    INCREMENTAL_LOAD_TYPE: type[UserIncrementalLoad] = UserIncrementalLoad

    def __init__(self, host: str, token: str) -> None:
        super().__init__(host, token)
        self.upstream_user_cache: dict[UserId, UserModel] = {}

    def _get_current_user_id(self) -> str:
        """Gets the current user ID."""

        profile_response = self._api.get_profile()

        if not profile_response.ok:
            raise Exception("Failed to get current user profile")

        profile_json = profile_response.json()
        profile = UserProfile.model_validate(profile_json)

        return profile.user_id

    def _try_get_user(
        self, user: UserModel, model: type[UserModel]
    ) -> UserModel | None:
        try:
            if user.user_id in self.upstream_user_cache:
                return self.upstream_user_cache[user.user_id]

            user_sdk_obj = self._api._sdk.catalog_user.get_user(user.user_id)
            return model.from_sdk_obj(user_sdk_obj)
        except NotFoundException:
            return None

    def _get_or_create_user_groups(self, groups: list[str]) -> None:
        """Ensures that all user groups exist in the project."""
        for group in groups:
            try:
                self._api._sdk.catalog_user.get_user_group(group)
            except NotFoundException:
                #  Create the user gtoup if it does not exist
                self._api.create_or_update_user_group(
                    CatalogUserGroup.init(
                        user_group_id=group, user_group_name=group
                    ),
                )
                self.logger.info(f"Created user group: {group}")

    def _user_is_equal_upstream(
        self,
        user: UserModel,
        upstream_user: UserModel | None,
    ) -> bool:
        """
        Checks if the user is different from the upstream user. Lists are checked by converting to sets.
        """
        if not upstream_user:
            return False

        user_data = user.model_dump()
        upstream_data = upstream_user.model_dump()

        for attr, source_value in user_data.items():
            upstream_value = upstream_data.get(attr)

            if isinstance(source_value, list):
                if set(source_value) != set(upstream_value or []):
                    return False
            else:
                if source_value != upstream_value:
                    return False
        return True

    def _create_or_update_user(
        self, user: UserModel, model: type[UserModel]
    ) -> None:
        """Creates or updates user in the project.

        Determines if the user needs to be updated or created by getting the
        upstream user from GoodData Cloud and comparing it with the source user.
        If user is supposed to be placed in a User Group, the function will check
        for its existence and create it if needed.

        """

        if user.user_id == self.current_user_id:
            self.logger.warning(
                f"Skipping creation/update of current user: {user.user_id}. "
                + "Current user should not be modified.",
            )
            return

        user_context = UserContext(
            user_id=user.user_id,
            user_groups=user.user_groups,
        )

        upstream_user = self._try_get_user(user, model)

        if self._user_is_equal_upstream(user, upstream_user):
            return

        self._get_or_create_user_groups(user.user_groups)

        self._api.create_or_update_user(
            user.to_sdk_obj(), **user_context.__dict__
        )
        self.logger.info(f"User {user.user_id} created/updated successfully.")

    def _delete_user(self, user_id: str) -> None:
        """Deletes user from the project."""
        if user_id == self.current_user_id:
            self.logger.warning(
                f"Skipping deletion of current user: {user_id}."
                + " Current user should not be deleted.",
            )
            return

        try:
            self._api._sdk.catalog_user.get_user(user_id)
        except NotFoundException:
            return

        self._api.delete_user(user_id)
        self.logger.info(f"Deleted user: {user_id}")

    def _manage_user(self, user: UserIncrementalLoad) -> None:
        """Manages user based on the provided GDUserTarget."""
        if user.is_active:
            self._create_or_update_user(user, UserIncrementalLoad)
        else:
            self._delete_user(user.user_id)

    def _provision_incremental_load(self) -> None:
        """Runs the incremental provisioning logic."""
        # Set the current user ID
        self.current_user_id = self._get_current_user_id()

        for user in self.source_group_incremental:
            # Attempt to process each user. On failure, log the error and continue
            try:
                self._manage_user(user)
            except Exception as e:
                self.logger.error(
                    f"Failed to manage user {user.user_id}. Error: {e} Context: {user.__dict__}"
                )

    def _provision_full_load(self) -> None:
        """Runs the full load provisioning logic."""

        # Set the current user ID
        self.current_user_id = self._get_current_user_id()

        # Get all upstream users
        catalog_upstream_users: list[CatalogUser] = self._api.list_users()

        # Convert catalog users to user models
        upstream_users: list[UserFullLoad] = [
            UserFullLoad.from_sdk_obj(user) for user in catalog_upstream_users
        ]

        # Cache the upstream users in a dict. It will be reused in `_try_get_user`
        self.upstream_user_cache = {
            user.user_id: user for user in upstream_users
        }
        # Get source IDs
        source_ids: set[str] = {user.user_id for user in self.source_group_full}

        # Get upstream IDs
        upstream_ids: set[str] = {user.user_id for user in upstream_users}

        # Create groups of IDs to delete, create, and in both systems
        id_groups = self._create_groups(source_ids, upstream_ids)

        # Iterate over source users and create/update
        for user in self.source_group_full:
            user_id = user.user_id

            if (
                user_id in id_groups.ids_to_create
                or user_id in id_groups.ids_in_both_systems
            ):
                self._create_or_update_user(user, UserFullLoad)

        # Delete users marked for deletion
        for user_id in id_groups.ids_to_delete:
            self._delete_user(user_id)



# (C) 2025 GoodData Corporation

from .entities.users.models.permissions import (
    PermissionFullLoad,
    PermissionIncrementalLoad,
)
from .entities.users.models.user_groups import (
    UserGroupFullLoad,
    UserGroupIncrementalLoad,
)
from .entities.users.models.users import (
    UserFullLoad,
    UserIncrementalLoad,
)
from .entities.users.permissions import PermissionProvisioner
from .entities.users.user_groups import UserGroupProvisioner
from .entities.users.users import UserProvisioner
from .entities.workspaces.workspace import WorkspaceProvisioner

__all__ = [
    "PermissionFullLoad",
    "PermissionIncrementalLoad",
    "PermissionProvisioner",
    "UserFullLoad",
    "UserGroupFullLoad",
    "UserIncrementalLoad",
    "UserGroupIncrementalLoad",
    "UserGroupProvisioner",
    "UserProvisioner",
    "WorkspaceProvisioner",
]



# (C) 2025 GoodData Corporation

from enum import Enum
from typing import Type, TypeAlias

import attrs

from gooddata_pipelines.provisioning.entities.users.models.permissions import (
    PermissionFullLoad,
    PermissionIncrementalLoad,
)
from gooddata_pipelines.provisioning.entities.users.models.user_groups import (
    UserGroupFullLoad,
    UserGroupIncrementalLoad,
)
from gooddata_pipelines.provisioning.entities.users.models.users import (
    UserFullLoad,
    UserIncrementalLoad,
)
from gooddata_pipelines.provisioning.entities.users.permissions import (
    PermissionProvisioner,
)
from gooddata_pipelines.provisioning.entities.users.user_groups import (
    UserGroupProvisioner,
)
from gooddata_pipelines.provisioning.entities.users.users import UserProvisioner
from gooddata_pipelines.provisioning.entities.workspaces.models import (
    WorkspaceFullLoad,
    WorkspaceIncrementalLoad,
)
from gooddata_pipelines.provisioning.entities.workspaces.workspace import (
    WorkspaceProvisioner,
)

ValidationModel: TypeAlias = (
    PermissionFullLoad
    | PermissionIncrementalLoad
    | UserFullLoad
    | UserIncrementalLoad
    | UserGroupFullLoad
    | UserGroupIncrementalLoad
    | WorkspaceFullLoad
    | WorkspaceIncrementalLoad
)

Provisioner: TypeAlias = (
    PermissionProvisioner
    | UserProvisioner
    | UserGroupProvisioner
    | WorkspaceProvisioner
)


class LoadType(str, Enum):
    FULL = "full"
    INCREMENTAL = "incremental"


class WorkflowType(str, Enum):
    WORKSPACE_FULL_LOAD = "workspace_full_load"
    WORKSPACE_INCREMENTAL_LOAD = "workspace_incremental_load"
    USER_FULL_LOAD = "user_full_load"
    USER_INCREMENTAL_LOAD = "user_incremental_load"
    USER_GROUP_FULL_LOAD = "user_group_full_load"
    USER_GROUP_INCREMENTAL_LOAD = "user_group_incremental_load"
    PERMISSION_FULL_LOAD = "permission_full_load"
    PERMISSION_INCREMENTAL_LOAD = "permission_incremental_load"


@attrs.define
class ProvisioningConfig:
    validation_model: Type[ValidationModel]
    provisioner_class: Type[Provisioner]
    load_type: LoadType


PROVISIONING_CONFIG = {
    WorkflowType.WORKSPACE_FULL_LOAD: ProvisioningConfig(
        validation_model=WorkspaceFullLoad,
        provisioner_class=WorkspaceProvisioner,
        load_type=LoadType.FULL,
    ),
    WorkflowType.WORKSPACE_INCREMENTAL_LOAD: ProvisioningConfig(
        validation_model=WorkspaceIncrementalLoad,
        provisioner_class=WorkspaceProvisioner,
        load_type=LoadType.INCREMENTAL,
    ),
    WorkflowType.USER_FULL_LOAD: ProvisioningConfig(
        validation_model=UserFullLoad,
        provisioner_class=UserProvisioner,
        load_type=LoadType.FULL,
    ),
    WorkflowType.USER_INCREMENTAL_LOAD: ProvisioningConfig(
        validation_model=UserIncrementalLoad,
        provisioner_class=UserProvisioner,
        load_type=LoadType.INCREMENTAL,
    ),
    WorkflowType.USER_GROUP_FULL_LOAD: ProvisioningConfig(
        validation_model=UserGroupFullLoad,
        provisioner_class=UserGroupProvisioner,
        load_type=LoadType.FULL,
    ),
    WorkflowType.USER_GROUP_INCREMENTAL_LOAD: ProvisioningConfig(
        validation_model=UserGroupIncrementalLoad,
        provisioner_class=UserGroupProvisioner,
        load_type=LoadType.INCREMENTAL,
    ),
    WorkflowType.PERMISSION_FULL_LOAD: ProvisioningConfig(
        validation_model=PermissionFullLoad,
        provisioner_class=PermissionProvisioner,
        load_type=LoadType.FULL,
    ),
    WorkflowType.PERMISSION_INCREMENTAL_LOAD: ProvisioningConfig(
        validation_model=PermissionIncrementalLoad,
        provisioner_class=PermissionProvisioner,
        load_type=LoadType.INCREMENTAL,
    ),
}



# (C) 2025 GoodData Corporation

"""Module for managing workspace data filter settings in GoodData Cloud."""

import json
from typing import Any
from uuid import uuid4

from requests import Response

from gooddata_pipelines.api import GoodDataApi
from gooddata_pipelines.logger.logger import LogObserver
from gooddata_pipelines.provisioning.entities.workspaces.models import (
    WDFSetting,
    WorkspaceDataMaps,
)
from gooddata_pipelines.provisioning.utils.context_objects import (
    WorkspaceContext,
)
from gooddata_pipelines.provisioning.utils.exceptions import WorkspaceException


class WorkspaceDataFilterManager:
    """
    Helper class to manage workspace data filter settings. Note that Workspace
    Data Filters themselves are not managed here. The Workspace Data Filter
    Setting object represents the relationship of values in a WDF column and
    a specific workspace.
    """

    def __init__(self, api: GoodDataApi, maps: WorkspaceDataMaps) -> None:
        self.api: GoodDataApi = api
        self.maps: WorkspaceDataMaps = maps
        self.logger: LogObserver = LogObserver()

    @staticmethod
    def _create_wdf_setting_dict(
        wdf_setting_id: str, wdf_id: str, wdf_values: list[str]
    ) -> dict[str, Any]:
        """Loads a JSON template of a WDF setting and fills it with the given values."""
        values = [str(value) for value in wdf_values]

        import os

        wdf_setting_path = os.path.join(
            os.path.dirname(__file__), "../../assets/wdf_setting.json"
        )
        with open(os.path.abspath(wdf_setting_path)) as file:
            wdf_setting: dict[str, Any] = json.load(file)

        wdf_setting["data"]["attributes"]["filterValues"] = values
        wdf_setting["data"]["id"] = wdf_setting_id
        wdf_setting["data"]["relationships"]["workspaceDataFilter"]["data"][
            "id"
        ] = wdf_id

        return wdf_setting

    def _get_wdf_settings_for_workspace(
        self, workspace_id: str
    ) -> list[WDFSetting]:
        """Gets all workspace data filter settings for a given workspace."""
        wdf_settings_response: Response = (
            self.api.get_workspace_data_filter_settings(workspace_id)
        )

        if not wdf_settings_response.ok:
            raise WorkspaceException(
                f"Failed to get WDF settings: {wdf_settings_response.text}",
                workspace_id=workspace_id,
                http_status=str(wdf_settings_response.status_code),
            )

        raw_wdf_settings: dict[str, Any] = wdf_settings_response.json()

        data: list[dict[str, Any]] = raw_wdf_settings["data"]

        wdf_settings: list[WDFSetting] = [
            WDFSetting(**wdf_setting) for wdf_setting in data
        ]

        return wdf_settings

    @staticmethod
    def _get_actual_wdf_setting_id_and_values(
        actual_wdf_settings: list[WDFSetting], actual_wdf_id: str
    ) -> tuple[str, list[str]]:
        """Returns WDF setting ID and values for given WDF ID."""
        for actual_wdf_setting in actual_wdf_settings:
            if (
                actual_wdf_setting.relationships.workspaceDataFilter["data"].id
                == actual_wdf_id
            ):
                actual_wdf_setting_id = actual_wdf_setting.id
                actual_wdf_values = actual_wdf_setting.attributes.filterValues

                return actual_wdf_setting_id, actual_wdf_values

        raise WorkspaceException(
            "Could not find WDF setting for WDF in actual WDF settings.",
            wdf_id=actual_wdf_id,
        )

    def _delete_redundant_wdf_setting(
        self,
        workspace_context: WorkspaceContext,
        actual_wdf_id: str,
        actual_wdf_settings: list[WDFSetting],
    ) -> None:
        """Deletes a WDF setting."""
        actual_wdf_setting_id, actual_wdf_values = (
            self._get_actual_wdf_setting_id_and_values(
                actual_wdf_settings, actual_wdf_id
            )
        )
        # Update context with actual values
        workspace_context.wdf_id = actual_wdf_id
        workspace_context.wdf_values = actual_wdf_values

        # If there is a WDF setting for a WDF that should not be associated with
        # the workspace, then delete the setting
        delete_response: Response = (
            self.api.delete_workspace_data_filter_setting(
                workspace_context.workspace_id,
                actual_wdf_setting_id,
            )
        )
        if delete_response.ok:
            self.logger.info(
                f"Deleted WDF setting for WDF {workspace_context.wdf_id} in "
                + f"workspace {workspace_context.workspace_id}"
            )
        else:
            raise WorkspaceException(
                f"Failed to delete WDF setting: {delete_response.text}",
                delete_response,
                workspace_context,
            )

    def _post_wdf_setting(
        self,
        workspace_context: WorkspaceContext,
    ) -> None:
        """Posts a WDF setting to Panther."""
        wdf_setting = self._create_wdf_setting_dict(
            str(uuid4()),
            str(workspace_context.wdf_id),
            workspace_context.wdf_values
            if workspace_context.wdf_values
            else [],
        )
        post_response: Response = self.api.post_workspace_data_filter_setting(
            workspace_context.workspace_id,
            wdf_setting,
        )
        if post_response.ok:
            self.logger.info(
                f"Created WDF setting for WDF {workspace_context.wdf_id} in workspace {workspace_context.workspace_id}"
            )
        else:
            raise WorkspaceException(
                f"Failed to create WDF setting: {post_response.text}",
                post_response,
                workspace_context,
            )

    def _put_wdf_setting(
        self,
        workspace_context: WorkspaceContext,
        actual_wdf_settings: list[WDFSetting],
    ) -> None:
        # get Panther WDF setting ID
        actual_wdf_setting_id, _ = self._get_actual_wdf_setting_id_and_values(
            actual_wdf_settings, str(workspace_context.wdf_id)
        )

        wdf_setting = self._create_wdf_setting_dict(
            actual_wdf_setting_id,
            str(workspace_context.wdf_id),
            workspace_context.wdf_values
            if workspace_context.wdf_values
            else [],
        )

        put_response: Response = self.api.put_workspace_data_filter_setting(
            workspace_context.workspace_id,
            wdf_setting,
        )
        if put_response.ok:
            self.logger.info(
                f"Updated WDF setting for WDF {workspace_context.wdf_id} in workspace {workspace_context.workspace_id}"
            )
        else:
            raise WorkspaceException(
                f"Failed to update WDF setting: {put_response.text}",
                put_response,
                workspace_context,
            )

    def _compare_wdf_settings(
        self,
        workspace_context: WorkspaceContext,
        source_wdf_config: dict[str, list[str]],
        upstream_wdf_settings: list[WDFSetting],
    ) -> None:
        """
        Compares WDF settings as extracted from the source with the actual WDF
        settings in Panther. We do not know the WDF setting IDs from the outset,
        which is why we need to check the WDF IDs and then the settings values
        in a roundabout way. I.e., we know that a WDF should have some setting
        with an unknown ID, but certain values. In this case, we don't care about
        the setting ID, but need to make sure that the workspace has the correct
        values for the WDF.
        """
        upstream_wdf_ids: set[str] = {
            upstream_wdf_setting.relationships.workspaceDataFilter["data"].id
            for upstream_wdf_setting in upstream_wdf_settings
        }

        source_wdf_ids: set[str] = set(source_wdf_config.keys())

        # Create map of upstream WDF_ID : WDF values
        upstream_wdf_ids_and_values: dict[str, list[str]] = {}
        for upstream_wdf_setting in upstream_wdf_settings:
            upstream_wdf_ids_and_values[
                upstream_wdf_setting.relationships.workspaceDataFilter[
                    "data"
                ].id
            ] = upstream_wdf_setting.attributes.filterValues

        # Iterate through source WDF settings
        for source_wdf_id in source_wdf_ids:
            # Update WDF information in context to make sure we have the correct
            # data -> there can be multiple WDFs per workspace
            source_values: list[str] = source_wdf_config[source_wdf_id]
            workspace_context.wdf_id = source_wdf_id
            workspace_context.wdf_values = source_values

            # Post WDF setting if missing for a WDF, when it should be there
            if source_wdf_id not in upstream_wdf_ids:
                self._post_wdf_setting(workspace_context)

            # If settings exist for a WDF that should be there, then compare values
            elif source_wdf_id in upstream_wdf_ids:
                actual_values: list[str] = upstream_wdf_ids_and_values[
                    source_wdf_id
                ]

                # If values are different, then update the WDF settings
                if set(source_values) != set(actual_values):
                    self._put_wdf_setting(
                        workspace_context,
                        upstream_wdf_settings,
                    )

        # Go through Panther WDF settings and check if there are any that should
        # not be there. Delete them if so.
        for actual_wdf_id in upstream_wdf_ids:
            if actual_wdf_id not in source_wdf_ids:
                self._delete_redundant_wdf_setting(
                    workspace_context,
                    actual_wdf_id,
                    upstream_wdf_settings,
                )

    def check_wdf_settings(
        self,
        workspace_context: WorkspaceContext,
    ) -> None:
        """
        Checks WDF settings for a given workspace.
        Creates WDF settings defined in source if they are missing in Panther.
        Updates WDF setting values if they are different in source and Panther.
        Deletes WDF settings from Panther if they are not defined in source.
        """
        actual_wdf_settings: list[WDFSetting] = (
            self._get_wdf_settings_for_workspace(workspace_context.workspace_id)
        )

        source_wdf_config: dict[str, list[str]] = (
            self.maps.workspace_id_to_wdf_map[workspace_context.workspace_id]
        )

        self._compare_wdf_settings(
            workspace_context, source_wdf_config, actual_wdf_settings
        )



# (C) 2025 GoodData Corporation

from pydantic import (
    BaseModel,
    ConfigDict,
)
from pydantic.alias_generators import (
    to_camel,
)


class Page(BaseModel):
    size: int
    total_elements: int
    total_pages: int
    number: int

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
    )


class Hierarchy(BaseModel):
    children_count: int

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
    )


class Meta(BaseModel):
    page: Page | None = None
    hierarchy: Hierarchy | None = None


class Workspace(BaseModel):
    id: str
    meta: Meta | None = None


class Links(BaseModel):
    self: str
    next: str | None = None


class WorkspaceResponse(BaseModel):
    data: list[Workspace]
    links: Links
    meta: Meta



# (C) 2025 GoodData Corporation

import csv
from typing import Iterator


class CSVReader:
    """Class to read the input CSV file and return its content as a list of strings."""

    @staticmethod
    def read_backup_csv(file_path: str) -> list[str]:
        """Reads the input CSV file, validates its structure, and returns its
        content as a list of strings.
        """

        with open(file_path) as csv_file:
            reader: Iterator[list[str]] = csv.reader(
                csv_file, skipinitialspace=True
            )

            try:
                # Skip the header
                headers = next(reader)

                if len(headers) > 1:
                    raise ValueError(
                        "Input file contains more than one column. Please check the input and try again."
                    )

            except StopIteration:
                # Raise an error if the iterator is empty
                raise ValueError("No content found in the CSV file.")

            # Read the content
            content = [row[0] for row in reader]

            # If the content is empty (no rows), raise an error
            if not content:
                raise ValueError("No workspaces found in the CSV file.")

        return content



# (C) 2025 GoodData Corporation
import datetime

import attrs
from gooddata_sdk._version import __version__ as sdk_version


@attrs.frozen
class DirNames:
    """
    Folder names used in the SDK backup process:
    - LAYOUTS - GoodData Layouts
    - LDM - Logical Data Model
    - AM - Analytics Model
    - UDF - User Data Filters
    """

    LAYOUTS = "gooddata_layouts"
    LDM = "ldm"
    AM = "analytics_model"
    UDF = "user_data_filters"


@attrs.frozen
class ApiDefaults:
    PAGE_SIZE = 100
    BATCH_SIZE = 100
    CALLS_PER_SECOND = 1.0


@attrs.frozen
class BackupSettings:
    API = ApiDefaults()
    MAX_RETRIES = 3
    RETRY_DELAY = 5  # seconds
    TIMESTAMP_SDK_FOLDER = (
        str(datetime.datetime.now().strftime("%Y%m%d-%H%M%S"))
        + "-"
        + sdk_version.replace(".", "_")
    )



# (C) 2025 GoodData Corporation

import abc
from pathlib import Path
from typing import Type, TypeVar

from gooddata_sdk.utils import PROFILES_FILE_PATH, profile_content

from gooddata_pipelines.api.gooddata_api_wrapper import GoodDataApi
from gooddata_pipelines.backup_and_restore.models.storage import (
    BackupRestoreConfig,
    StorageType,
)
from gooddata_pipelines.backup_and_restore.storage.base_storage import (
    BackupStorage,
)
from gooddata_pipelines.backup_and_restore.storage.local_storage import (
    LocalStorage,
)
from gooddata_pipelines.backup_and_restore.storage.s3_storage import S3Storage
from gooddata_pipelines.logger import LogObserver
from gooddata_pipelines.utils.file_utils import JsonUtils, YamlUtils

ManagerT = TypeVar("ManagerT", bound="BaseManager")


class BaseManager(abc.ABC):
    """Base class to provide constructors for backup and restore managers."""

    storage: BackupStorage

    def __init__(self, host: str, token: str, config: BackupRestoreConfig):
        self.config = config

        self._api: GoodDataApi = GoodDataApi(host, token)
        self.logger: LogObserver = LogObserver()

        self.storage = self._get_storage(self.config)

        self.yaml_utils = YamlUtils()
        self.json_utils = JsonUtils()

    def _get_storage(self, conf: BackupRestoreConfig) -> BackupStorage:
        """Returns the storage class based on the storage type."""
        if conf.storage_type == StorageType.S3:
            return S3Storage(conf)
        elif conf.storage_type == StorageType.LOCAL:
            return LocalStorage(conf)
        else:
            raise RuntimeError(
                f'Unsupported storage type "{conf.storage_type.value}".'
            )

    @classmethod
    def create(
        cls: Type[ManagerT],
        config: BackupRestoreConfig,
        host: str,
        token: str,
    ) -> ManagerT:
        """Creates a backup worker instance using the provided host and token."""
        return cls(host=host, token=token, config=config)

    @classmethod
    def create_from_profile(
        cls: Type[ManagerT],
        config: BackupRestoreConfig,
        profile: str = "default",
        profiles_path: Path = PROFILES_FILE_PATH,
    ) -> ManagerT:
        """Creates a backup worker instance using a GoodData profile file."""
        content = profile_content(profile, profiles_path)
        return cls(host=content["host"], token=content["token"], config=config)



# (C) 2025 GoodData Corporation
import pytest
from pydantic import ValidationError

from gooddata_pipelines.ldm_extension.models.custom_data_object import (
    ColumnDataType,
    CustomDataset,
    CustomDatasetDefinition,
    CustomFieldDefinition,
    CustomFieldType,
)


def make_valid_field_def(**kwargs):
    data = {
        "workspace_id": "ws1",
        "dataset_id": "ds1",
        "custom_field_id": "cf1",
        "custom_field_name": "Custom Field",
        "custom_field_type": CustomFieldType.ATTRIBUTE,
        "custom_field_source_column": "col1",
        "custom_field_source_column_data_type": ColumnDataType.STRING,
    }
    data.update(kwargs)
    return data


def make_valid_dataset_def(**kwargs):
    data = {
        "workspace_id": "ws1",
        "dataset_id": "ds1",
        "dataset_name": "Dataset",
        "dataset_datasource_id": "dsrc1",
        "dataset_source_table": "table1",
        "dataset_source_sql": None,
        "parent_dataset_reference": "parent_ds",
        "parent_dataset_reference_attribute_id": "parent_attr",
        "dataset_reference_source_column": "src_col",
        "dataset_reference_source_column_data_type": ColumnDataType.STRING,
        "workspace_data_filter_id": "wdf1",
        "workspace_data_filter_column_name": "col1",
    }
    data.update(kwargs)
    return data


def test_custom_field_definition_valid():
    field = CustomFieldDefinition(**make_valid_field_def())
    assert field.custom_field_id == "cf1"
    assert field.custom_field_type == CustomFieldType.ATTRIBUTE


def test_custom_field_definition_cf_id_equals_dataset_id_raises():
    data = make_valid_field_def(custom_field_id="ds1")
    with pytest.raises(ValidationError) as exc:
        CustomFieldDefinition(**data)
    assert "cannot be the same as dataset ID" in str(exc.value)


def test_custom_dataset_definition_valid_table():
    ds = CustomDatasetDefinition(**make_valid_dataset_def())
    assert ds.dataset_source_table == "table1"
    assert ds.dataset_source_sql is None


def test_custom_dataset_definition_valid_sql():
    data = make_valid_dataset_def(
        dataset_source_table=None, dataset_source_sql="SELECT 1"
    )
    ds = CustomDatasetDefinition(**data)
    assert ds.dataset_source_sql == "SELECT 1"
    assert ds.dataset_source_table is None


def test_custom_dataset_definition_both_none_raises():
    data = make_valid_dataset_def(
        dataset_source_table=None, dataset_source_sql=None
    )
    with pytest.raises(ValidationError) as exc:
        CustomDatasetDefinition(**data)
    assert "must be provided" in str(exc.value)


def test_custom_dataset_definition_both_provided_raises():
    data = make_valid_dataset_def(
        dataset_source_table="table1", dataset_source_sql="SELECT 1"
    )
    with pytest.raises(ValidationError) as exc:
        CustomDatasetDefinition(**data)
    assert (
        "Only one of dataset_source_table and dataset_source_sql can be provided"
        in str(exc.value)
    )


def test_custom_dataset_model():
    ds_def = CustomDatasetDefinition(**make_valid_dataset_def())
    field_def = CustomFieldDefinition(**make_valid_field_def())
    dataset = CustomDataset(definition=ds_def, custom_fields=[field_def])
    assert dataset.definition.dataset_id == "ds1"
    assert len(dataset.custom_fields) == 1
    assert dataset.custom_fields[0].custom_field_id == "cf1"



# (C) 2025 GoodData Corporation

"""Logging observer for the GoodData Pipelines SDK.

This module provides a singleton observer class `LogObserver` that allows
subscribing logger-like objects to receive log messages. The observer emits
unformatted log messages to all subscribed objects, which should implement
the `LoggerLike` protocol.
"""

from enum import Enum
from typing import Any, Protocol


class SingletonMeta(type):
    _instances: dict = {}

    def __call__(cls, *args: Any, **kwargs: Any) -> "SingletonMeta":
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


class Severity(Enum):
    """Severity levels for logging."""

    INFO = "info"
    WARNING = "warning"
    ERROR = "error"


class LoggerLike(Protocol):
    """A protocol for a logger-like object.

    This protocol defines the methods that a logger-like object should implement
    to be compatible with the `LogObserver`. It includes methods for logging
    messages at different severity levels: info, warning, and error.
    """

    def info(self, *args: Any, **kwargs: Any) -> None: ...

    def warning(self, *args: Any, **kwargs: Any) -> None: ...

    def error(self, *args: Any, **kwargs: Any) -> None: ...


class LogObserver(metaclass=SingletonMeta):
    """Singleton observer class for logging messages.

    Emits unformatted log messages to all subscribed logger-like objects.
    """

    # TODO: in future we might want to add a timestamp or other metadata
    # (severity...)? Currently that is left out to subscribers to handle.

    # TODO: with error we're dumping the context as string to the message
    # that could be improved (either passing the context as a separate arg
    # or handling the process here).

    def __init__(self) -> None:
        self.subscribers: list[LoggerLike] = []

    def subscribe(self, subscriber: LoggerLike) -> None:
        """Subscribe a logger-like object to receive log messages.

        Args:
            subscriber (LoggerLike): An object that implements the LoggerLike
            protocol.
        Returns:
            None
        """
        self.subscribers.append(subscriber)

    def unsubscribe(self, subscriber: LoggerLike) -> None:
        """Unsubscribe a logger-like object from receiving log messages.

        Args:
            subscriber (LoggerLike): An object that implements the LoggerLike
            protocol.

        Returns:
            None
        """
        self.subscribers.remove(subscriber)

    def _notify(self, severity: Severity, msg: str) -> None:
        """Notify all subscribers with a log message.

        Args:
            severity (Severity): The severity level of the log message.
            msg (str): The log message to be sent to subscribers.

        Returns:
            None
        """
        for subscriber in self.subscribers:
            if severity == Severity.INFO:
                subscriber.info(msg)
            elif severity == Severity.WARNING:
                subscriber.warning(msg)
            elif severity == Severity.ERROR:
                subscriber.error(msg)

    def info(self, msg: str) -> None:
        """Log an info message."""
        self._notify(Severity.INFO, msg)

    def warning(self, msg: str) -> None:
        """Log a warning message."""
        self._notify(Severity.WARNING, msg)

    def error(self, msg: str) -> None:
        """Log an error message."""
        self._notify(Severity.ERROR, msg)
